import DKIDE.Parse
import DKIDE.Project
import Data.List as L
import Data.Tree
import Data.Map.Lazy as ML
import Data.IORef
import Data.Text
import Data.Char
import System.IO as IO
import System.Process
import System.Directory
import System.Exit
import Graphics.UI.Gtk
import Graphics.UI.Gtk.SourceView
import Control.Monad.Trans
import Control.Monad
import Control.Concurrent
import Control.Concurrent.STM
import Control.Monad.STM
import Numeric
import Text.Read (readEither)
import Rose.Graphics.REPL
import Rose.Types
import Rose.Misc
import Rose.Buildins
import Debug.Trace

runFileDialog :: FileChooserAction -> Maybe Window -> IO String
runFileDialog mode parent =
    fileChooserDialogNew (case mode of
    	FileChooserActionOpen 				-> Just "Open file"
    	FileChooserActionSave 				-> Just "Save file"
    	FileChooserActionSelectFolder -> Just "Select folder"
    	FileChooserActionCreateFolder -> Just "Create folder") parent mode [("Ok",ResponseOk),("Cancel",ResponseCancel)] >>= (\dialog ->
        dialogRun dialog >>= (\resp ->
            case resp of
                ResponseOk -> fileChooserGetFilename dialog >>= (\x ->
                    case x of
                        Just v -> widgetDestroy dialog >> return v
                        _ -> widgetDestroy dialog >> return "")
                _ -> widgetDestroy dialog >> return ""))

runNameRequest :: Builder -> String -> IO String
runNameRequest builder title =
	builderGetObject builder castToDialog "nameDialog" >>= (\dialog ->
		dialog `set` [windowTitle := title] >>
		dialogSetDefaultResponse dialog (ResponseUser 1) >>
		dialogRun dialog >>= (\resp ->
			case resp of
				ResponseUser 0 ->
					builderGetObject builder castToEntry "fNameInput" >>= (\entry ->
						widgetHide dialog >>
						entrySetText entry "" >>
						return "")
				ResponseUser 1 ->
					builderGetObject builder castToEntry "fNameInput" >>= (\entry ->
						entryGetText entry >>= (\txt ->
							widgetHide dialog >>
							entrySetText entry "" >>
							return txt))))

runErrorDialog :: Builder -> String -> String -> IO ()
runErrorDialog builder title message =
	builderGetObject builder castToMessageDialog "errorDialog" >>= (\errorDialog ->
		errorDialog `set` [messageDialogText := Just title, messageDialogSecondaryText := Just message] >>
		dialogRun errorDialog >>
		widgetHide errorDialog)

runHelpDialog :: Builder -> IO ()
runHelpDialog builder =
	builderGetObject builder castToDialog "helpDialog" >>= (\helpDialog ->
		dialogRun helpDialog >>
		widgetHide helpDialog)

runYesNoDialog :: Builder -> String -> String -> IO Bool
runYesNoDialog builder title body =
	builderGetObject builder castToDialog "yesNoDialog" >>= (\dialog ->
		dialog `set` [windowTitle := title] >>
		builderGetObject builder castToLabel "yesNoLabel" >>= (\label ->
			label `set` [labelLabel := body] >>
			dialogRun dialog >>= (\r ->
				widgetHide dialog >>
				case r of
					ResponseUser 0 -> return False
					ResponseUser 1 -> return True)))

initTreeViewTreeStore :: TreeView -> IO (TreeStore String)
initTreeViewTreeStore tv =
	treeStoreNew [] >>= (\ts ->
		tv `set` [treeViewModel := ts] >>
		treeViewColumnNew >>= (\tc ->
			treeViewColumnSetTitle tc "Project files" >>
			treeViewAppendColumn tv tc >>
			cellRendererTextNew >>= (\crt ->
				treeViewColumnPackStart tc crt True >>
				cellLayoutSetAttributes tc crt ts (\s -> [cellText := s]))) >>
				return ts)

initTreeViewListStore :: TreeView -> IO (ListStore String)
initTreeViewListStore tv =
	listStoreNew [] >>= (\ts ->
		tv `set` [treeViewModel := ts] >>
		treeViewColumnNew >>= (\tc ->
			treeViewColumnSetTitle tc "Calc variables" >>
			treeViewAppendColumn tv tc >>
			cellRendererTextNew >>= (\crt ->
				treeViewColumnPackStart tc crt True >>
				cellLayoutSetAttributes tc crt ts (\s -> [cellText := s]))) >>
				return ts)

initComboBoxListStore :: ComboBox -> IO (ListStore ComboBoxText)
initComboBoxListStore cb =
	comboBoxSetModelText cb

initCentralSourceView :: IO (SourceView,SourceBuffer)
initCentralSourceView =
		sbuffer >>= (\sb ->
			sourceViewNewWithBuffer sb >>= (\sw ->
				sw `set` [sourceViewHighlightCurrentLine 	:= True,
									sourceViewShowLineNumbers 			:= True,
									sourceViewTabWidth 							:= 8,
									sourceViewAutoIndent 						:= True] >>
				return (sw,sb)))
	where
		sbuffer =
			sourceLanguageManagerNew >>= (\smanager ->
				sourceLanguageManagerGetLanguage smanager "cpp" >>= (\sl ->
					case sl of
						Just fsl -> sourceBufferNewWithLanguage fsl
						Nothing -> error "Nebyl nalezen jazyk cpp!"))

expandedRows = "expandedRows"
currentPrimaryTab = "currentPrimaryTab"

roseStdLibPath = "stdlib.rose"
exaBuildPath = "exabuild.rose"

graphicFilePath = "graphic.glade"

createStartFiles :: String -> IO ()
createStartFiles path =
	let
		base = path
		projname = nameFromPath path
		realpath = base ++ "/" ++ projname ++ ".dkide"
	in
		IO.openFile realpath WriteMode >>= (\h ->
			IO.hPutStr h (show (
				Project projname base [projname ++ ".ldscript","entry.s","kmain.cpp","build.rose"] [] [2] ["-std=c++11","-O0"] [] []
				(ML.fromList [(expandedRows,"[True]"),(currentPrimaryTab,"0")]))) >>
			IO.hClose h) >>
		IO.openFile (base ++ "/" ++ projname ++ ".ldscript") WriteMode >>= (\h ->
			IO.hClose h) >>
		IO.openFile (base ++ "/" ++ "entry.s") WriteMode >>= (\h ->
			IO.hClose h) >>
		IO.openFile (base ++ "/" ++ "kmain.cpp") WriteMode >>= (\h ->
			IO.hClose h) >>
		IO.openFile (base ++ "/" ++ "build.rose") WriteMode >>= (\h -> do
			ebc <- IO.readFile exaBuildPath
			IO.hPutStr h ebc
			IO.hClose h)

saveCurrentTab :: CurrentProject -> Builder -> IO ()
saveCurrentTab project builder =
	builderGetObject builder castToNotebook "sourceNoteBook" >>= (\sourceNoteBook ->
		sourceNoteBook `get` notebookPage >>= (\page ->
			modifyCurrentProject (\p -> modifyIdeArgs (\m -> ML.insert currentPrimaryTab (show page) m) p) project))

setCurrentTab :: CurrentProject -> Builder -> IO ()
setCurrentTab project builder =
	builderGetObject builder castToNotebook "sourceNoteBook" >>= (\sourceNoteBook ->
		loadCurrentProject project >>= (\p ->
			sourceNoteBook `set` [notebookPage := read (ideArgs p ! currentPrimaryTab)]))

saveCurrentFile :: CurrentProject -> SourceBuffer -> IO ()
saveCurrentFile proj buff =
	loadCurrentProject proj >>= (\p ->
		let
			cpath = getFilePath (0:(currentOpenFile p)) p
		in
			buff `get` textBufferText >>= (\t ->
				IO.writeFile cpath t))

loadFile :: CurrentProject -> TreePath -> SourceBuffer -> IO ()
loadFile proj path buff =
	loadCurrentProject proj >>= (\p ->
		let
			tpath = getFilePath path p
		in
			buff `set` [textBufferText :=> IO.readFile tpath])

refreshSourceEdit :: CurrentProject -> SourceView -> SourceBuffer -> Label -> IO ()
refreshSourceEdit proj view buff label =
	loadCurrentProject proj >>= (\p ->
		let
			fpath = getFilePath (0:(currentOpenFile p)) p
			fname = nameFromPath fpath
			isLibraryFile = L.length (currentOpenFile p) == 2
		in
			if isLibraryFile
			then
				buff `set` [textBufferText :=> IO.readFile fpath] >>
				label `set` [labelLabel := fname] >>
				view `set` [textViewEditable := False]
			else
				buff `set` [textBufferText :=> IO.readFile fpath] >>
				label `set` [labelLabel := fname] >>
				view `set` [textViewEditable := True])

refreshProjectTree :: CurrentProject -> TreeView -> TreeStore String -> IO ()
refreshProjectTree proj tview tstore =
		treeStoreClear tstore >>
		loadCurrentProject proj >>= (\p ->
			let
				(root:expandTable) = read (ideArgs p ! expandedRows) :: [Bool]
				lengthPrefix = L.length (projectFiles p)
			in
				treeStoreInsertTree tstore [] 0 (getFullProjectTree p) >>
				(if root
				then treeViewExpandRow tview [0] False
				else treeViewCollapseRow tview [0]) >>
				applyETable (\i e ->
					if e
					then
						treeViewExpandRow  tview [0,lengthPrefix + i] False
					else
						treeViewCollapseRow tview [0,lengthPrefix + i]) expandTable 0)
	where
		applyETable _ [] _ =
			return ()
		applyETable fun (x:xs) c =
			fun c x >> applyETable fun xs (c + 1)

refreshSourceEditAndProjectTree :: CurrentProject -> SourceView -> SourceBuffer -> Label -> TreeView -> TreeStore String -> IO ()
refreshSourceEditAndProjectTree proj view buff label tview tstore =
	refreshSourceEdit proj view buff label >>
	refreshProjectTree proj tview tstore

consoleReplInit :: CurrentProject -> Builder -> SourceBuffer -> SourceView -> TVar [String] -> IO ()
consoleReplInit project builder sbuff sview sfiles = do
	view <- builderGetObject builder castToTextView "runReport"
	buff <- builderGetObject builder castToTextBuffer "runReportBuffer"
	runGraphicsRepl view buff sfiles
		"For DKIDE help, type dk-help(). Good luck and happy coding."
		["dk-help","dk-path-sep","dk-base-dir","dk-proj-name",
		"dk-lib-files","dk-proj-files","dk-config","dk-compile",
		"dk-link","dk-run","dk-make-lib-symlinks","dk-destroy-lib-symlinks",
		"dk-save-current-file","dk-tab-width","dk-gpp-args","dk-ld-args"]
		(stdBuildins `mergeBuildins` buildins [
		("dk-tab-width",\input ->
			case input of
				[] -> do
					idd <- sview `get` sourceViewTabWidth
					return (Right (RInt idd))
				[v] ->
					case v of
						RInt i -> do
							sview `set` [sourceViewTabWidth := i]
							return (Right RUnit)
						_ ->
							return (Left (badType "dk-tab-width" (deepValType v) "Int" "first argument"))
				_ ->
					return (Left (badArgCount "dk-tab-width" (L.length input) 1))),
		("dk-gpp-args",\input ->
			case input of
				[] -> do
					p <- loadCurrentProject project
					return (Right (RList (L.map (\a -> RList (L.map (\x -> RChar x) a)) (gppArgs p))))
				[v] ->
					case v of
						RList ilist -> do
							let gargs = L.map unpackString ilist
							let isok = L.foldl (\a x -> if x == Nothing then False else a) True gargs
							let unpackit = L.map (\(Just x) -> x) gargs
							if isok
							then do
								modifyCurrentProject (\p -> modifyGppArgs (\_ -> unpackit) p) project
								return (Right RUnit)
							else
								return (Left (badType "dk-gpp-args" (deepValType v) "[[Char]]" "first argument"))
						_ ->
							return (Left (badType "dk-gpp-args" (deepValType v) "[[Char]]" "first argument"))
				_ ->
					return (Left (badArgCount "dk-gpp-args" (L.length input) 1))),
		("dk-ld-args",\input ->
			case input of
				[] -> do
					p <- loadCurrentProject project
					return (Right (RList (L.map (\a -> RList (L.map (\x -> RChar x) a)) (ldArgs p))))
				[v] ->
					case v of
						RList ilist -> do
							let gargs = L.map unpackString ilist
							let isok = L.foldl (\a x -> if x == Nothing then False else a) True gargs
							let unpackit = L.map (\(Just x) -> x) gargs
							if isok
							then do
								modifyCurrentProject (\p -> modifyLdArgs (\_ -> unpackit) p) project
								return (Right RUnit)
							else
								return (Left (badType "dk-ld-args" (deepValType v) "[[Char]]" "first argument"))
						_ ->
							return (Left (badType "dk-ld-args" (deepValType v) "[[Char]]" "first argument"))
				_ ->
					return (Left (badArgCount "dk-ld-args" (L.length input) 1))),
		("dk-help",\_ ->
			return (Right RUnit)),
		("dk-path-sep",\_ ->
			return (Right (RList [RChar '/']))),
		("dk-base-dir",\_ -> do
			p <- loadCurrentProject project
			return (Right (RList (L.map (\c -> RChar c) (projectPath p))))),
		("dk-proj-name",\_ -> do
			p <- loadCurrentProject project
			return (Right (RList (L.map (\c -> RChar c) (projectName p))))),
		("dk-lib-files",\_ -> do
			p <- loadCurrentProject project
			let lfiles =
				L.concat (L.map (\(Node b fss) -> L.map (\(Node f _) -> b ++ "/" ++ f) fss) (libraryFiles p))
			return (Right (RList (L.map (\f -> RList (L.map (\c -> RChar c) f)) lfiles)))),
		("dk-proj-files",\_ -> do
			p <- loadCurrentProject project
			return (Right (RList (L.map (\s -> RList (L.map (\c -> RChar c) (projectPath p ++ "/" ++ s))) (projectFiles p))))),
		-- dk-compile(file-path) -> (report-string || _ if bad file type, obj-path || _ if fail)
		("dk-compile",\input -> do
			p <- loadCurrentProject project
			case input of
				[spath'] ->
					case unpackString spath' of
						Just spath -> do
							let sname = nameFromPath spath
							let oname = removeDotType (nameFromPath spath) ++ ".o"
							let stype = getDotType sname
							let pbase = projectPath p
							let gppargs = L.concat (L.intersperse " " (gppArgs p))
							case stype of
								"s" -> do
									(Just sin,Just sout,Just serr,handle) <- createProcess (CreateProcess (ShellCommand ("as " ++ spath ++ " -o " ++ pbase ++
										"/" ++ oname)) (Just (purePath spath)) Nothing CreatePipe CreatePipe CreatePipe False False)
									ecode <- waitForProcess handle
									case ecode of
										ExitSuccess -> do
											report <- hGetContents sout
											return (Right (RTuple [RList (L.map (\c -> RChar c) report),RList (L.map (\c -> RChar c) (pbase ++ "/" ++ oname))]))
										ExitFailure ec -> do
											ereport <- hGetContents serr
											return (Right (RTuple [RList (L.map (\c -> RChar c) ("AS failed with " ++ show ec ++ ": " ++ ereport)),RUnit]))
								"cpp" -> do
									(Just sin,Just sout,Just serr,handle) <- createProcess (CreateProcess (ShellCommand ("g++ " ++ gppargs ++
										" -ffreestanding -nostdlib -w -c " ++ spath ++ " -o " ++ pbase ++ "/" ++ oname)) (Just (purePath spath)) Nothing
										CreatePipe CreatePipe CreatePipe False False)
									ecode <- waitForProcess handle
									case ecode of
										ExitSuccess -> do
											report <- hGetContents sout
											return (Right (RTuple [RList (L.map (\c -> RChar c) report),RList (L.map (\c -> RChar c) (pbase ++ "/" ++ oname))]))
										ExitFailure ec -> do
											ereport <- hGetContents serr
											return (Right (RTuple [RList (L.map (\c -> RChar c) ("G++ failed with " ++ show ec ++ ": " ++ ereport)),RUnit]))
								_ -> return (Right (RTuple [RUnit,RUnit]))
						Nothing -> return (Left (badType "dk-compile" (deepValType spath') "[Char]" "first argument"))
				_ -> return (Left (badArgCount "dk-compile" (L.length input) 1))),
		-- dk-link(bin-name,linker-script,[object-files]) -> (report-string,bin-path || _ if fail)
		("dk-link",\input -> do
			p <- loadCurrentProject project
			case input of
				[bname',lname',ofpaths'] ->
					case unpackString bname' of
						Just bname ->
							case unpackString lname' of
								Just lname ->
									if valType ofpaths' == "RList"
									then
										let
											RList ilist = ofpaths'
											msources = L.map unpackString ilist
											isfail = L.foldl (\a ms -> case ms of Nothing -> Just 0;_ -> a) Nothing msources
											ofpaths = L.map (\(Just x) -> x) msources
										in
											case isfail of
												Just _ -> return (Left (badType "dk-link" (deepValType ofpaths') "[[Char]]" "second argument"))
												_ ->
													let
														lscript = projectPath p ++ "/" ++ lname
														ofiles = L.concat (L.intersperse " " ofpaths)
														ldargs = L.concat (L.intersperse " " (ldArgs p))
														base = projectPath p
														shcomm =
															"ld " ++ ldargs ++ " -T " ++ lscript ++ " " ++ ofiles ++ " -o " ++ base ++ "/" ++ bname
													in do
														(Just sin,Just sout,Just serr,handle) <- createProcess (CreateProcess (ShellCommand shcomm)
															(Just base) Nothing CreatePipe CreatePipe CreatePipe False False)
														ecode <- waitForProcess handle
														case ecode of
															ExitSuccess -> do
																report <- hGetContents sout
																return (Right (RTuple [RList (L.map (\c -> RChar c) report),
																	RList (L.map (\c -> RChar c) (base ++ "/" ++ bname))]))
															ExitFailure ec -> do
																ereport <- hGetContents serr
																return (Right (RTuple [RList (L.map (\c -> RChar c) ("LD failed with " ++ show ec ++ ": " ++ ereport)),RUnit]))
									else return (Left (badType "dk-link" (deepValType ofpaths') "[[Char]]" "third argument"))
								Nothing -> return (Left (badType "dk-link" (deepValType lname') "[Char]" "second argument"))
						Nothing -> return (Left (badType "dk-link" (deepValType bname') "[Char]" "first argument"))
				_ -> return (Left (badArgCount "dk-link" (L.length input) 3))),
		-- dk-run(fda-path,fdb-path,hda-path,hdb-path,hdc-path,hdd-path,cdrom-path,m-memcount,boot-option)
		-- -boot c
		-- -m 64
		-- -hda image
		("dk-run",\input -> do
			p <- loadCurrentProject project
			if L.length input == 9
			then
				let
					minput = L.map unpackString input
					isfail = L.foldl (\a ms -> case ms of Nothing -> Just 0;_ -> a) Nothing minput
					uinput = L.map (\(Just x) -> x) minput
					fdaopt = 	if (uinput !! 0) == " " then "" else "-fda " ++ 	uinput !! 0
					fdbopt = 	if (uinput !! 1) == " " then "" else "-fdb " ++ 	uinput !! 1
					hdaopt = 	if (uinput !! 2) == " " then "" else "-hda " ++ 	uinput !! 2
					hdbopt = 	if (uinput !! 3) == " " then "" else "-hdb " ++ 	uinput !! 3
					hdcopt = 	if (uinput !! 4) == " " then "" else "-hdc " ++ 	uinput !! 4
					hddopt =  if (uinput !! 5) == " " then "" else "-hdd " ++ 	uinput !! 5
					cdromopt =if (uinput !! 6) == " " then "" else "-cdrom " ++ uinput !! 6
					mopt = 		if (uinput !! 7) == " " then "" else "-m " ++ 		uinput !! 7
					bopt = 		if (uinput !! 8) == " " then "" else "-boot " ++ 	uinput !! 8
					qargs = L.concat (L.intersperse " " (L.filter (/="") [fdaopt,hdaopt,hdbopt,hdcopt,fdbopt,hddopt,cdromopt,mopt,bopt]))
				in
					case isfail of
						Just _ -> return (Left (badType "dk-run" (fullListTypes input) "[[Char]]" "arguments"))
						_ -> do
							(_,_,_,handle) <- createProcess (CreateProcess (ShellCommand ("xterm -e qemu-system-x86_64 -monitor stdio " ++ qargs)) (Just "/")
								Nothing CreatePipe CreatePipe CreatePipe False False)
							waitForProcess handle
							return (Right RUnit)
			else return (Left (badArgCount "dk-run" (L.length input) 9))),
		("dk-make-lib-symlinks",\_ -> do
			p <- loadCurrentProject project
			mapM_ (\libpath -> do
				(sin,sout,serr,handle) <- createProcess (shell ("ln -s " ++ libpath ++ " " ++ (projectPath p ++ "/" ++ nameFromPath libpath)))
				waitForProcess handle) (L.map (\(Node n _) -> n) (libraryFiles p))
			return (Right RUnit)),
		("dk-destroy-lib-symlinks",\_ -> do
			p <- loadCurrentProject project
			let libs = L.map (\(Node n _) -> n) (libraryFiles p)
			let	libnames = L.map nameFromPath libs
			let	base = projectPath p
			let	links = L.map (\l -> base ++ "/" ++ l) libnames
			mapM_ (\l -> removeFile l) links
			return (Right RUnit)),
		("dk-save-current-file",\input -> do
			saveCurrentFile project sbuff
			return (Right RUnit))]) (saveCurrentFile project sbuff)

minorTerminationHandler :: CurrentProject -> SourceBuffer -> TextBuffer -> Builder -> IO ()
minorTerminationHandler proj buff rrbuff builder =
	saveCurrentFile proj buff >>
	saveCurrentTab proj builder >>
	mainQuit

majorTerminationHandler :: CurrentProject -> Window -> SourceBuffer -> TextBuffer -> Builder -> IO ()
majorTerminationHandler proj win buff rrbuff builder =
	saveCurrentFile proj buff >>
	saveCurrentTab proj builder >>
	widgetDestroy win >>
	mainQuit

projectRowActivatedHandler :: TreeView -> TreeStore String -> CurrentProject -> SourceBuffer -> SourceView
															-> Label -> TreePath -> TreeViewColumn -> IO ()
projectRowActivatedHandler tv tvstore proj buff _ label path@(_:[]) _ =
	loadCurrentProject proj >>= (\p ->
		let
			etable = read (ideArgs p ! expandedRows) :: [Bool]
			expand = etable !! 0
			newetable e = e : L.tail etable
		in
			if expand
			then
				modifyCurrentProject (modifyIdeArgs (\m -> ML.update (\_ ->
					Just (show (newetable False))) expandedRows m)) proj >>
				refreshProjectTree proj tv tvstore
			else
				modifyCurrentProject (modifyIdeArgs (\m -> ML.update (\_ ->
					Just (show (newetable True))) expandedRows m)) proj) >>
				refreshProjectTree proj tv tvstore
projectRowActivatedHandler tv _ proj buff view label path@(_:x:[]) _ =
	loadCurrentProject proj >>= (\p ->
		if L.length (projectFiles p) <= x
		then
			let
				lenprefix = L.length (projectFiles p)
				indexToTable = x - lenprefix
				etable = read (ideArgs p ! expandedRows) :: [Bool]
				expand = etable !! (indexToTable + 1)
				newetable e = L.take (indexToTable + 1) etable ++ e : L.drop (indexToTable + 2) etable
			in
				if expand
				then
					treeViewCollapseRow tv [0,x] >>
					modifyCurrentProject (modifyIdeArgs (\m -> ML.update (\_ ->
						Just (show (newetable False))) expandedRows m)) proj
				else
					treeViewExpandRow tv [0,x] False >>
					modifyCurrentProject (modifyIdeArgs (\m -> ML.update (\_ ->
						Just (show (newetable True))) expandedRows m)) proj
		else
			saveCurrentFile proj buff >>
			modifyCurrentProject (\(Project pn pp pf lf opfile g l q i) ->
				Project pn pp pf lf (L.tail path) g l q i) proj >>
			refreshSourceEdit proj view buff label)
projectRowActivatedHandler _ _ proj buff view label (_:x:y:[]) _ =
	loadCurrentProject proj >>= (\p ->
		saveCurrentFile proj buff >>
		modifyCurrentProject (\(Project pn pp pf lf opfile g l q i) ->
			Project pn pp pf lf [x - L.length (projectFiles p),y] g l q i) proj >>
		refreshSourceEdit proj view buff label)

projectTreeEventClickHandler builder project projectTree =
	eventTime >>= (\etime ->
		eventButton >>= (\ebutton ->
			if ebutton == RightButton
			then
				liftIO (treeViewGetCursor projectTree >>= (\(tpath,_) ->
					loadCurrentProject project >>= (\p ->
						let
							pathlen = L.length tpath
							projfilenum = L.length (projectFiles p)
						in
							if pathlen == 0
							then
								return True
							else if pathlen == 1
							then
								builderGetObject builder castToMenu "rootProjectPopup" >>= (\rootProjectPopup ->
									menuPopup rootProjectPopup (Just (RightButton,etime))) >> return True
							else if pathlen == 2
							then
								if (tpath !! 1) < projfilenum
								then
									builderGetObject builder castToMenu "projectFilePopup" >>= (\projectFilePopup ->
										menuPopup projectFilePopup (Just (RightButton,etime))) >>
										builderGetObject builder castToMenuItem "pfpRemoveProjectFile" >>= (\m ->
											treeViewGetCursor projectTree >>= (\(path,_) ->
												loadCurrentProject project >>= (\p ->
													let
														targetFile = nameFromPath (getFilePath path p)
													in
														m `set` [menuItemLabel := ("Remove project file " ++ targetFile)]))) >>
										return True
								else
									builderGetObject builder castToMenu "projectLibraryPopup" >>= (\projectLibraryPopup ->
										menuPopup projectLibraryPopup (Just (RightButton,etime))) >>
										builderGetObject builder castToMenuItem "plpRemoveProjectLibrary" >>= (\m ->
											treeViewGetCursor projectTree >>= (\(path,_) ->
												loadCurrentProject project >>= (\p ->
													let
														targetLib = nameFromPath (getFilePath path p)
													in
														m `set` [menuItemLabel := ("Remove project libary " ++ targetLib)]))) >>
										return True
							else return True)))
				else
					liftIO (return False)))

newProjectHandler :: Builder -> Window -> CurrentProject -> SourceBuffer -> SourceView -> Label -> TreeView
										 -> TreeStore String -> TextBuffer -> IO ()
newProjectHandler builder window project sourceBuffer sourceEdit sourceLabel projectTree projectTreeStore rrbuff =
	runYesNoDialog builder "New project" "Do you really want new project? Don't worry, the old one will be saved." >>= (\r ->
		if r == False
		then return ()
		else
			saveCurrentFile project sourceBuffer >>
			saveCurrentTab project builder >>
			runFileDialog FileChooserActionSave (Just window) >>= (\ppath ->
				if ppath == []
				then return ()
				else
					createStartFiles ppath >>
					replaceCurrentProject (ppath ++ "/" ++ (nameFromPath ppath) ++ ".dkide") project >>
					refreshSourceEditAndProjectTree project sourceEdit sourceBuffer sourceLabel projectTree projectTreeStore >>
					setCurrentTab project builder))

openProjectHandler :: Builder -> Window -> CurrentProject -> SourceBuffer -> SourceView -> Label -> TreeView
									 		-> TreeStore String -> TextBuffer -> IO ()
openProjectHandler builder window project sourceBuffer sourceEdit sourceLabel projectTree projectTreeStore rrbuff =
	runYesNoDialog builder "Open project" "Do you really want open new project? Don't worry, the old one will be saved." >>= (\r ->
		if r == False
		then return ()
		else
			saveCurrentFile project sourceBuffer >>
			saveCurrentTab project builder >>
			runFileDialog FileChooserActionOpen (Just window) >>= (\ppath ->
				if ppath == []
				then return ()
				else
					initCurrentProject "" >>= (\np ->
						replaceCurrentProject ppath np >>= (\r ->
							case r of
								[] ->
									consistenciCheck np >>= (\(cr,cp) ->
										showConsistenciReport builder cr cp (
											loadCurrentProject np >>= (\np' ->
												modifyCurrentProject (\_ -> np') project) >>
											refreshSourceEditAndProjectTree project sourceEdit sourceBuffer sourceLabel projectTree projectTreeStore >>
											setCurrentTab project builder) False)
								err ->
									runErrorDialog builder "Neplatný soubor"
									("Soubor " ++ ppath ++ " nebylo možné otevřít jako konfigurační soubor projektu DKIDE." ++
									" Buď otvíráte špatný soubor a nebo jste ručně a špatně upravovali obsah konfiguračního souboru projektu! " ++
									"Parse report: " ++ err)))))

addFileHandler :: Window -> CurrentProject -> TreeView -> TreeStore String -> Builder -> TVar [String] -> IO ()
addFileHandler window project tview tstore builder lfiles =
	runNameRequest builder "File name" >>= (\txt ->
		if txt == []
		then return ()
		else
			loadCurrentProject project >>= (\p ->
				if txt `L.elem` projectFiles p
				then
					runErrorDialog builder "Duplicitní jméno" ("Soubor s jménem " ++ txt ++ " je již v projektu!")
				else do
					IO.openFile (projectPath p ++ "/" ++ txt) WriteMode >>= (\h -> IO.hClose h)
					modifyCurrentProject (\p' -> addProjectFile txt p') project
					refreshProjectTree project tview tstore
					if getDotType txt == "rose"
					then do
						atomically (modifyTVar lfiles (\x -> (projectPath p ++ "/" ++ txt) : x))
					else return ()))

removeFileHandler :: String -> CurrentProject -> SourceView -> SourceBuffer -> Label -> TreeView
	-> TreeStore String -> Builder -> TVar [String] -> IO ()
removeFileHandler fname project sview sbuff label tview tstore builder lfiles =
		runYesNoDialog builder "Remove project file" "Do you really want remove project file?" >>= (\r ->
			if r == False
			then return ()
			else
				loadCurrentProject project >>= (\p ->
					if fname == []
					then return ()
					else
						if fname `L.elem` [projectName p ++ ".ldscript","entry.s","kmain.cpp","build.rose"]
						then
							runErrorDialog builder "Zamknutý soubor" ("Pokoušíte se smazat zamknutý soubor " ++ fname ++
							". Ani jeden z kořenových souborů projektu není možné smazat z důvodu udržení konzistence. Nesmazatelné soubory jsou " ++
							projectName p ++ ".ldscript" ++ ", entry.s, build.rose a kmain.cpp!")
						else
							if fname `L.elem` projectFiles p
							then do
								removeFile (projectPath p ++ "/" ++ fname)
								modifyCurrentProject (trans fname) project
								refreshSourceEditAndProjectTree project sview sbuff label tview tstore
								atomically (modifyTVar lfiles (\x -> L.filter (\x' -> (projectPath p ++ "/" ++ fname) /= x') x))
							else
								runErrorDialog builder "Neexistující soubor"
								("V projektu (knihovní soubory není možné upravovat) nebyl nalezen soubor " ++ fname ++ "!")))
	where
		trans file (Project pn pp pf lf (x:[]) g l q i) =
				Project pn pp (L.filter (/=file) pf) lf (if indexOfFile <= x then [x - 1] else [x]) g l q i
			where
				(Just indexOfFile) = elemIndex file pf
		trans file (Project pn pp pf lf (x:y:[]) g l q i) =
				Project pn pp (L.filter (/=file) pf) lf [x,y] g l q i

addLibraryHandler :: Window -> CurrentProject -> SourceView -> SourceBuffer -> Label -> TreeView -> TreeStore String -> Builder -> IO ()
addLibraryHandler window project sview sbuff label tview tstore builder =
	loadCurrentProject project >>= (\p ->
		runFileDialog FileChooserActionSelectFolder (Just window) >>= (\path ->
			if path == []
			then return ()
			else
				if L.foldl (\found (Node n _) -> if found then found else nameFromPath n == nameFromPath path) False (libraryFiles p)
				then
					runErrorDialog builder "Duplicitní knihovna" ("Není možné importovat knihovnu " ++ path ++ " protože jej iž importována!")
				else
					getDirectoryContents path >>= (\files ->
						let
							wantedFiles = L.filter (\f -> getDotType f `L.elem` ["s","mac","cpp","h"]) files
							libtree = Node path (L.map (\f -> Node f []) wantedFiles)
						in
							modifyCurrentProject (addLibrary libtree) project >>
							modifyCurrentProject (\(Project pn pp pf lf opfile g l q i) ->
								let
									newmap = ML.update (\v -> Just (show ((read v :: [Bool]) ++ [False]))) expandedRows i
								in
									Project pn pp pf lf opfile g l q newmap) project >>
							refreshProjectTree project tview tstore)))

removeLibraryHandler :: String -> CurrentProject -> SourceView -> SourceBuffer -> Label -> TreeView -> TreeStore String -> Builder -> IO ()
removeLibraryHandler name project sview sbuff label tview tstore builder =
		runYesNoDialog builder "Remove the libary" "Do you really want remove the library? Don't worry, physical files are not touched." >>= (\r ->
			if r == False
			then return ()
			else
				loadCurrentProject project >>= (\p ->
					if name == []
					then return ()
					else
						if L.foldl (\found (Node n _) -> if found then found else nameFromPath n == name) False (libraryFiles p)
						then
							let
								etable = read (ideArgs p ! expandedRows) :: [Bool]
								Just libindex = L.elemIndex name (L.map (\(Node n _) -> nameFromPath n) (libraryFiles p))
								newetable = L.take libindex etable ++ L.drop (libindex+1) etable
							in
								saveCurrentFile project sbuff >>
								modifyCurrentProject (trans name) project >>
								modifyCurrentProject (modifyIdeArgs (\m -> ML.update (\_ -> Just (show newetable)) expandedRows m)) project >>
								refreshSourceEditAndProjectTree project sview sbuff label tview tstore
						else
							runErrorDialog builder "Neexistující knihovna" ("Knihovna " ++ name ++ " v projektu importována není!")))
	where
		trans name p@(Project pn pp pf lf (x:[]) g l q i) =
			removeLibrary name p
		trans name p@(Project pn pp pf lf (x:y:[]) g l q i) =
			modifyCurrentOpenFile (\_ -> [2]) (removeLibrary name p)

aboutHandler :: Builder -> IO ()
aboutHandler builder =
	builderGetObject builder castToAboutDialog "dkideAbout" >>= (\dkideAbout ->
		dialogRun dkideAbout >>
		widgetHide dkideAbout)

dkideMain project = do
	p'' <- loadCurrentProject project
	sfiles <- atomically (newTVar ([roseStdLibPath] ++ L.filter (/=[]) (L.map (\f ->
		if getDotType f == "rose"
		then projectPath p'' ++ "/" ++ f
		else []) (projectFiles p''))))
	builderNew >>= (\builder ->
		builderAddFromFile builder "graphic.glade" >>
		builderGetObject builder castToWindow "mainWindow" >>= (\window ->
			builderGetObject builder castToScrolledWindow "sourceViewScrollWindow" >>= (\sourceScrollWindow ->
				builderGetObject builder castToLabel "sourceLabel" >>= (\sourceLabel ->
					builderGetObject builder castToTextView "runReport" >>= (\runReport ->
						builderGetObject builder castToTextBuffer "runReportBuffer" >>= (\runReportBuffer ->
							builderGetObject builder castToTreeView "projectTree" >>= (\projectTree ->
								initCentralSourceView >>= (\(sourceEdit,sourceBuffer) ->
									initTreeViewTreeStore projectTree >>= (\projectTreeStore ->
										sourceScrollWindow `set` [containerChild := sourceEdit] >>
										loadCurrentProject project >>= (\p -> window `set` [windowTitle := projectName p]) >>
										refreshSourceEditAndProjectTree project sourceEdit sourceBuffer sourceLabel projectTree projectTreeStore >>
										setCurrentTab project builder >>
										consoleReplInit project builder sourceBuffer sourceEdit sfiles >>
										widgetShowAll window >>
										windowMaximize window >>
										on window unrealize (minorTerminationHandler project sourceBuffer runReportBuffer builder) >>
										on projectTree rowActivated (projectRowActivatedHandler projectTree projectTreeStore project
											sourceBuffer sourceEdit sourceLabel) >>
										on projectTree buttonPressEvent (projectTreeEventClickHandler builder project projectTree) >>
										on runReport sizeAllocate (\_ ->
											textBufferGetEndIter runReportBuffer >>= (\it ->
												textViewScrollToIter runReport it 0 Nothing >> return ())) >>
										builderGetObject builder castToMenuItem "newProject" >>= (\newProject ->
											on newProject menuItemActivated (newProjectHandler builder window project sourceBuffer sourceEdit
											sourceLabel projectTree projectTreeStore runReportBuffer)) >>
										builderGetObject builder castToMenuItem "openProject" >>= (\openProject ->
											on openProject menuItemActivated (openProjectHandler builder window project sourceBuffer sourceEdit
											sourceLabel projectTree projectTreeStore runReportBuffer)) >>
										builderGetObject builder castToMenuItem "pfpRemoveProjectFile" >>= (\pfpRemoveProjectFile ->
											on pfpRemoveProjectFile menuItemActivated (
												treeViewGetCursor projectTree >>= (\(path,_) ->
													loadCurrentProject project >>= (\p ->
														let
															targetFile = nameFromPath (getFilePath path p)
														in
															removeFileHandler targetFile project sourceEdit sourceBuffer
															sourceLabel projectTree projectTreeStore builder sfiles)))) >>
										builderGetObject builder castToMenuItem "plpRemoveProjectLibrary" >>= (\plpRemoveProjectLibrary ->
											on plpRemoveProjectLibrary menuItemActivated (
												treeViewGetCursor projectTree >>= (\(path,_) ->
													loadCurrentProject project >>= (\p ->
														let
															targetLib = nameFromPath (getFilePath path p)
														in
															removeLibraryHandler targetLib project sourceEdit sourceBuffer
															sourceLabel projectTree projectTreeStore builder)))) >>
										builderGetObject builder castToMenuItem "rppAddProjectLibrary" >>= (\rppAddProjectLibrary ->
											on rppAddProjectLibrary menuItemActivated (addLibraryHandler window project sourceEdit
											sourceBuffer sourceLabel projectTree projectTreeStore builder)) >>
										builderGetObject builder castToMenuItem "rppAddProjectFile" >>= (\rppAddProjectFile ->
											on rppAddProjectFile menuItemActivated (addFileHandler window project projectTree projectTreeStore builder sfiles)) >>
										return ())))))))))

showConsistenciReport :: Builder -> String -> Bool -> IO () -> Bool -> IO ()
showConsistenciReport builder report canContinue operation haltExecution = do
	consistenciReportDialog <- builderGetObject builder castToDialog "consistenciReportDialog"
	crdTVBuffer <- builderGetObject builder castToTextBuffer "crdTVBuffer"
	crdOK <- builderGetObject builder castToButton "crdOK"
	if report == []
	then
		operation
	else do
		crdTVBuffer `set` [textBufferText := report]
		widgetShowAll consistenciReportDialog
		if canContinue
		then
			on crdOK buttonActivated (widgetHide consistenciReportDialog >> operation) >> return ()
		else if haltExecution
		then do
			on crdOK buttonActivated (do
				widgetDestroy consistenciReportDialog
				mainQuit)
			return ()
		else do
			on crdOK buttonActivated (widgetHide consistenciReportDialog)
			return ()

consistenciCheck :: CurrentProject -> IO (String,Bool)
consistenciCheck project = do
	p <- loadCurrentProject project
	let pname = projectName p
	let pbase = projectPath p
	let	pfiles = L.map (\f -> pbase ++ "/" ++ f) (projectFiles p)
	let	lfiles = L.concat (L.map (\(Node lpref childs) -> L.map (\(Node f _) -> lpref ++ "/" ++ f) childs) (libraryFiles p))
	let	afiles = pfiles ++ lfiles
	let	testFilesForExistence = do
		mapM (\f -> do
			e <- doesFileExist f
			return (f,e)) afiles
	f <- testFilesForExistence
	let gfiles =
		[pbase ++ "/" ++ pname ++ ".ldscript",pbase ++ "/" ++ "entry.s",pbase ++ "/" ++ "kmain.cpp",pbase ++ "/" ++ "build.rose"]
	let missingGuarantedFiles =
		L.foldl (\a (xv,xp) ->
			if xp
			then L.filter (/=xv) a
			else a) gfiles f
	let missingFiles =
		L.filter (\xv -> xv `L.notElem` gfiles) (L.map (\(xv,_) -> xv) (L.filter (\(_,xp) -> xp == False) f))
	let createMissingGFiles =
		foldM (\a cf -> do
			h <- IO.openFile cf WriteMode
			if nameFromPath cf == "build.rose"
			then do
				ec <- IO.readFile exaBuildPath
				IO.hPutStr h ec
				IO.hClose h
			else
				IO.hClose h
			return (a ++ "Guaranted file " ++ cf ++ " not found ... problem solved via creation of default file!\n")) "" missingGuarantedFiles
	let removeMissingFilesFromProject =
		if missingFiles == []
		then return []
		else do
			modifyCurrentProject (\(Project pn pp pf lf opfile g l q i) ->
				let
					newpf =
						L.filter (\x -> (pp ++ "/" ++ x) `L.notElem` missingFiles) pf
					newlf =
						L.map (\(Node lpref lfiles) -> Node lpref (L.filter (\(Node x _) -> lpref ++ "/" ++ x `L.notElem` missingFiles) lfiles)) lf
				in do
					Project pn pp newpf newlf [2] g l q i) project
			return (L.concat (L.map (\mf ->
				"File " ++ mf ++ " not found ... problem solved via its deletion from project!\n") missingFiles))
	let testIDEArgs =
		let
			iargs = ideArgs p
			iatest = L.map (\ta -> (ta,ML.member ta iargs)) [expandedRows]
		in do
			modifyCurrentProject (\pr -> modifyIdeArgs (\im -> L.foldl (\a (an,ap) ->
				if ap
				then a
				else ML.insert an (if an == expandedRows
					then ("[True,False,False,False,False,False,False,False,False,False,False]")
					else error ("RUNTIME CRITICAL ERROR 101: Unknown missing arg " ++ an ++ "!")) a) im iatest) pr) project
			return (L.concat (L.map (\(av,ap) ->
				if ap == False
				then "IDE arg " ++ av ++ " not found in project ide args ... problem solved via its default generation!\n"
				else []) iatest))
	r0 <- createMissingGFiles
	r1 <-	removeMissingFilesFromProject
	r2 <- testIDEArgs
	return (r0 ++ r1 ++ r2,True)

dkideOpenProject = do
	ppath <- runFileDialog FileChooserActionOpen Nothing
	if ppath == []
	then
		mainQuit
	else do
		cp <- initCurrentProject ppath
		r <- replaceCurrentProject ppath cp
		case r of
			[] -> do
				(r,p) <- consistenciCheck cp
				if r == []
				then
					dkideMain cp
				else do
					builder <- builderNew
					builderAddFromFile builder graphicFilePath
					showConsistenciReport builder r p (dkideMain cp) True
			err -> do
				builder <- builderNew
				builderAddFromFile builder graphicFilePath
				runErrorDialog builder "Neplatný soubor" ("Soubor " ++ ppath ++
					" nebylo možné otevřít jako konfigurační soubor projektu DKIDE." ++
					" Buď otvíráte špatný soubor a nebo jste ručně a špatně upravovali obsah konfiguračního souboru projektu! " ++
					"Parse report: " ++ err)
				mainQuit

dkideNewProject = do
	ppath <- runFileDialog FileChooserActionCreateFolder Nothing
	if ppath == []
	then
		mainQuit
	else do
		createStartFiles ppath
		cp <- initCurrentProject ppath
		replaceCurrentProject (ppath ++ "/" ++ (nameFromPath ppath) ++ ".dkide") cp
		dkideMain cp

dkideInitWindow = do
		(gpp,binutils,qemu,terminal,all) <- testForWare
		builder <- builderNew
		builderAddFromFile builder graphicFilePath
		dialog <-	builderGetObject builder castToDialog "initDialog"
		widgetShowAll dialog
		cancelButton <-	builderGetObject builder castToButton "cancelButton"
		on cancelButton buttonActivated (do
			widgetDestroy dialog
			mainQuit)
		gppstat <- builderGetObject builder castToLabel "gppStatus"
		binstat <- builderGetObject builder castToLabel "binutilsStatus"
		qstat <- builderGetObject builder castToLabel "qemuStatus"
		tstat <- builderGetObject builder castToLabel "terminalStatus"
		gppstat `set` [labelLabel := if gpp then "<span foreground=\"white\" background=\"green\">Found</span>"
			else "<span foreground=\"white\" background=\"red\">Not found</span>", labelUseMarkup := True]
		binstat `set` [labelLabel := if binutils then "<span foreground=\"white\" background=\"green\">Found</span>"
			else "<span foreground=\"white\" background=\"red\">Not found</span>", labelUseMarkup := True]
		qstat `set` [labelLabel := if qemu then "<span foreground=\"white\" background=\"green\">Found</span>"
			else "<span foreground=\"white\" background=\"red\">Not found</span>", labelUseMarkup := True]
		tstat `set` [labelLabel := if terminal then "<span foreground=\"white\" background=\"green\">Found</span>"
			else "<span foreground=\"white\" background=\"red\">Not found</span>", labelUseMarkup := True]
		if all == False
		then do
			b <- builderGetObject builder castToVButtonBox "vbuttonbox1"
			widgetHideAll b
		else do
			opButton <- builderGetObject builder castToButton "openProjectButton"
			npButton <-	builderGetObject builder castToButton "newProjectButton"
			on opButton buttonActivated (do
				widgetDestroy dialog
				dkideOpenProject)
			on npButton buttonActivated (do
				widgetDestroy dialog
				dkideNewProject)
			return ()
	where
		testForWare = do
				g <- testG
				b <- testB
				q <- testQ
				t <- testT
				return (g,b,q,t,g && b && q && t)
			where
				testG = do
					(_,_,_,h) <- createProcess (CreateProcess (RawCommand "g++" ["--help"]) Nothing Nothing CreatePipe
						CreatePipe CreatePipe True False)
					ec <-	waitForProcess h
					case ec of
						ExitSuccess -> return True
						_ -> return False
				testB = do
					(_,_,_,h) <- createProcess (CreateProcess (RawCommand "ld" ["--help"]) Nothing Nothing CreatePipe
						CreatePipe CreatePipe True False)
					ec <-	waitForProcess h
					case ec of
						ExitSuccess -> return True
						_ -> return False
				testQ = do
					(_,_,_,h) <- createProcess (CreateProcess (RawCommand "qemu-system-i386" ["-h"]) Nothing Nothing CreatePipe
						CreatePipe CreatePipe True False)
					ec <-	waitForProcess h
					case ec of
						ExitSuccess -> return True
						_ -> return False
				testT = do
					(_,_,_,h) <- createProcess (CreateProcess (RawCommand "xterm" ["-h"]) Nothing Nothing CreatePipe
						CreatePipe CreatePipe True False)
					ec <-	waitForProcess h
					case ec of
						ExitSuccess -> return True
						_ -> return False

gtkmain = do
	initGUI
	dkideInitWindow
	mainGUI

main = gtkmain
