module Rose.Buildins (
	buildins,
	mergeBuildins,
	addBuildins,
	stdBuildins,
	) where

import Data.List as L
import Data.Map.Lazy as ML
import Data.Char
import System.IO as IO
import System.Directory
import Data.Char  (digitToInt)
import Data.Maybe (listToMaybe)
import Numeric    (readInt,showIntAtBase)

import Rose.Types
import Rose.Misc
import Rose.Serialize
import Rose.Parse

stdBuildins = 
	buildins [addBuildin,subBuildin,mulBuildin,divBuildin,
	remBuildin,tointBuildin,tofloatBuildin,toboolBuildin,
	tocharBuildin,andBuildin,orBuildin,notBuildin,
	equalBuildin,nequalBuildin,greaterBuildin,lesserBuildin,
	eqorgrBuildin,eqorleBuildin,headBuildin,tailBuildin,
	initBuildin,backBuildin,growBuildin,consBuildin,
	lenBuildin,atBuildin,memberBuildin,lookupBuildin,
	insertBuildin,removeBuildin,seqBuildin,repeatBuildin,
	putStrBuildin,putStrLnBuildin,readLnBuildin,
	readAllBuildin,tConstructBuildin,lConstructBuildin,
	mConstructBuildin,typeinfoBuildin,typecheckBuildin,
	errorBuildin,traceBuildin,writeFileBuildin,readFileBuildin,
	appendFilesBuildin,fileExistBuildin,deleteFilesBuildin,
	showBuildin,readBuildin,tobinBuildin,tohexBuildin,
	appendBuildin]

buildins :: 
	[(String,[RVal] -> IO (Either String RVal))] 
	-> ML.Map String ([RVal] -> IO (Either String RVal))
buildins = ML.fromList

mergeBuildins :: 
	ML.Map String ([RVal] -> IO (Either String RVal)) 
	-> ML.Map String ([RVal] -> IO (Either String RVal)) 
	-> ML.Map String ([RVal] -> IO (Either String RVal))
mergeBuildins = ounion

addBuildins :: 
	ML.Map String ([RVal] -> IO (Either String RVal))
	-> [(String,[RVal] -> IO (Either String RVal))] 
	-> ML.Map String ([RVal] -> IO (Either String RVal))
addBuildins b0 b1 =
	mergeBuildins b0 (buildins b1)
	
addBuildin :: (String,[RVal] -> IO (Either String RVal))
addBuildin =
	("add",\inputs ->
		case listType inputs of
			"RInt" -> 
				return (Right (RInt (L.foldl (+) 0 (unpackRInts inputs))))
			"RFloat" ->
				return (Right (RFloat (L.foldl (+) 0 (unpackRFloats inputs))))
			e -> return (Left (badType "add" (fullListTypes inputs) "Int or Float" "")))

subBuildin :: (String,[RVal] -> IO (Either String RVal))
subBuildin =
	("sub",\inputs ->
		case listType inputs of
			"RInt" -> 
				return (Right (RInt (L.foldl (-) (L.head (unpackRInts inputs)) (L.tail (unpackRInts inputs)))))
			"RFloat" ->
				return (Right (RFloat (L.foldl (-) (L.head (unpackRFloats inputs)) (L.tail (unpackRFloats inputs)))))
			e -> return (Left (badType "sub" (fullListTypes inputs) "Int or Float" "")))
			
mulBuildin :: (String,[RVal] -> IO (Either String RVal))
mulBuildin =
	("mul",\inputs ->
		case listType inputs of
			"RInt" -> 
				return (Right (RInt (L.foldl (*) 1 (unpackRInts inputs))))
			"RFloat" ->
				return (Right (RFloat (L.foldl (*) 1 (unpackRFloats inputs))))
			e -> return (Left (badType "mul" (fullListTypes inputs) "Int or Float" "")))
			
divBuildin :: (String,[RVal] -> IO (Either String RVal))
divBuildin =
	("div",\inputs ->
		case listType inputs of
			"RInt" -> 
				return (Right (RInt (L.foldl div (L.head $ unpackRInts inputs) (L.tail $ unpackRInts inputs))))
			"RFloat" ->
				return (Right (RFloat (L.foldl (/) (L.head $ unpackRFloats inputs) (L.tail $ unpackRFloats inputs))))
			e -> return (Left (badType "div" (fullListTypes inputs) "Int or Float" "")))
			
remBuildin :: (String,[RVal] -> IO (Either String RVal))
remBuildin =
	("rem",\inputs ->
		case listType inputs of
			"RInt" -> 
				return (Right (RInt (L.foldl rem (L.head $ unpackRInts inputs) (L.tail $ unpackRInts inputs))))
			e -> return (Left (badType "rem" (fullListTypes inputs) "Int or Float" "")))

tointBuildin :: (String,[RVal] -> IO (Either String RVal))
tointBuildin =
	("to-int",\input ->
		case input of
			[v] ->
				case valType v of
					"RInt" -> return (Right v) 
					"RFloat" -> return (Right (RInt (floor (unpackRFloat v))))
					"RChar" -> return (Right (RInt (ord (unpackRChar v))))
					"RBool" -> if unpackRBool v then return (Right (RInt 1)) else return (Right (RInt 0))
					"RList" ->
						let
							RList ilist = v
							istring = listType ilist == "RChar"
						in
							if istring 
							then
								case readMaybe (L.map (\(RChar x) -> x) ilist) of
									Just r -> return (Right (RInt r))
									_ -> 
										let
											readBin :: Integral a => String -> Maybe a
											readBin str = (fmap fst . listToMaybe . readInt 2 (`L.elem` "01") digitToInt) (L.drop 2 str) 
										in
											case readBin (L.map (\(RChar x) -> x) ilist) of
												Just r -> return (Right (RInt r))
												_ -> return (Right RUnit)
							else
								return (Left (badType "to-int" (listType ilist) "Char" "supplied list in first argument"))
					"" -> return (Right RUnit)
					err -> return (Left (badType "to-int" err "Int, Float, Char, Bool or [Char]" "in first argument"))
			binput -> return (Left (badArgCount "to-int" (L.length binput) 1)))
			
tofloatBuildin :: (String,[RVal] -> IO (Either String RVal))
tofloatBuildin =
	("to-float",\input ->
		case input of
			[v] ->	
				case valType v of
					"RInt" -> return (Right (RFloat (fromIntegral (unpackRInt v)))) 
					"RFloat" -> return (Right v)
					"RChar" -> return (Right (RFloat (fromIntegral (ord (unpackRChar v)))))
					"RBool" -> if unpackRBool v then return (Right (RFloat 1)) else return (Right (RFloat 0))
					"RList" ->
						let
							RList ilist = v
							istring = listType ilist == "RChar"
						in
							if istring 
							then
								case readMaybe (L.map (\(RChar x) -> x) ilist) of
									Just r -> return (Right (RFloat r))
									_ -> return (Right RUnit)
							else
								return (Left (badType "to-float" (listType ilist) "Char" "supplied list in first argument"))
					"" -> return (Right RUnit)
					err -> return (Left (badType "to-float" err "Int, Float, Char or Bool" ""))
			binput -> return (Left (badArgCount "to-float" (L.length binput) 1)))
			
toboolBuildin :: (String,[RVal] -> IO (Either String RVal))
toboolBuildin =
	("to-bool",\input ->
		case input of
			[v] ->		
				case valType v of
					"RInt" -> return (Right (if unpackRInt v == 0 then RBool False else RBool True)) 
					"RFloat" -> return (Right (if unpackRFloat v == 0 then RBool False else RBool True))
					"RBool" -> return (Right v)
					err -> return (Left (badType "to-bool" err "Int, Float or Bool" ""))
			binput -> return (Left (badArgCount "to-bool" (L.length binput) 1)))
			
tocharBuildin :: (String,[RVal] -> IO (Either String RVal))
tocharBuildin =
	("to-char",\input ->
		case input of
			[v] ->		
				case valType v of
					"RInt" -> return (Right (RChar (chr (unpackRInt v)))) 
					"RChar" -> return (Right v)
					err -> return (Left (badType "to-char" err "Int or Char" ""))
			binput -> return (Left (badArgCount "to-char" (L.length binput) 1)))

tobinBuildin :: (String,[RVal] -> IO (Either String RVal))
tobinBuildin =
	("to-bin",\input ->
		case input of
			[v] ->		
				case valType v of
					"RInt" -> return (Right (RList (L.map (\c -> RChar c) ("0b" ++ showIntAtBase 2 intToDigit (unpackRInt v) ""))))
					err -> return (Left (badType "to-bin" err "Int" "first argument"))
			binput -> return (Left (badArgCount "to-bin" (L.length binput) 1)))
			
tohexBuildin :: (String,[RVal] -> IO (Either String RVal))
tohexBuildin =
	("to-hex",\input ->
		case input of
			[v] ->		
				case valType v of
					"RInt" -> return (Right (RList (L.map (\c -> RChar c) ("0x" ++ showIntAtBase 16 intToDigit (unpackRInt v) ""))))
					err -> return (Left (badType "to-hex" err "Int" "first argument"))
			binput -> return (Left (badArgCount "to-hex" (L.length binput) 1)))

andBuildin :: (String,[RVal] -> IO (Either String RVal))
andBuildin =
	("and",\inputs ->
		case listType inputs of
			"RBool" -> return (Right (RBool (L.foldl (&&) (L.head (unpackRBools inputs)) (L.tail (unpackRBools inputs)))))
			err -> return (Left (badType "and" err "Bool" "")))
	
orBuildin :: (String,[RVal] -> IO (Either String RVal))
orBuildin =
	("or",\inputs ->
		case listType inputs of
			"RBool" -> return (Right (RBool (L.foldl (||) (L.head (unpackRBools inputs)) (L.tail (unpackRBools inputs)))))
			err -> return (Left (badType "or" err "Bool" "")))

notBuildin :: (String,[RVal] -> IO (Either String RVal))
notBuildin =
	("not",\input ->
		case input of
			[v] ->		
				case valType v of
					"RBool" -> return (Right (RBool (not (unpackRBool v))))
					err -> return (Left (badType "not" err "Bool" ""))
			binput -> return (Left (badArgCount "not" (L.length binput) 1)))

equalBuildin :: (String,[RVal] -> IO (Either String RVal))
equalBuildin =
	("equal",\input ->
		case input of
			[v0,v1] -> return (Right (RBool (v0 == v1)))
			binput -> return (Left (badArgCount "equal" (L.length binput) 2)))

nequalBuildin :: (String,[RVal] -> IO (Either String RVal))
nequalBuildin =
	("nequal",\input ->
		case input of
			[v0,v1] -> return (Right (RBool (v0 /= v1)))
			binput -> return (Left (badArgCount "nequal" (L.length binput) 2)))

ordBuildin :: String -> (String,[RVal] -> IO (Either String RVal))
ordBuildin optype =
	(optype,\input ->
		case input of
			[v0,v1] ->		
				if valType v0 == valType v1
				then
					case valType v0 of
						"RInt" -> 
							return (Right (RBool ((case optype of 
								"greater" -> (>)
								"lesser" -> (<)
								"eqorgr" -> (>=)
								"eqorle" -> (<=)) (unpackRInt v0) (unpackRInt v1))))
						"RFloat" -> 
							return (Right (RBool ((case optype of 
								"greater" -> (>)
								"lesser" -> (<)
								"eqorgr" -> (>=)
								"eqorle" -> (<=)) (unpackRFloat v0) (unpackRFloat v1))))
						"RChar" -> 
							return (Right (RBool ((case optype of 
								"greater" -> (>)
								"lesser" -> (<)
								"eqorgr" -> (>=)
								"eqorle" -> (<=)) (unpackRChar v0) (unpackRChar v1))))
						"RList" ->
							let
								(_,this) = ordBuildin optype
								RList v0list = v0
								RList v1list = v1
								res' = sequence (L.zipWith (\z0 z1 -> this [z0,z1]) v0list v1list)   
							in
								res' >>= (\res ->
									if findError res == ""
									then
										let
											unpackedRes = collectValues res
										in 
											return (Right (RBool (L.foldl (\a x -> 
												if a == unpackRBool x then a else False) True unpackedRes)))
									else return (Left ("Function " ++ optype ++ " failed in List compare. Error is: " ++ findError res)))
						err -> return (Left (badType optype err "Int, Float, Char or List" ""))
				else return (Left (badInput optype ("the both arguments to be the same type. " ++ valType v0 ++ "!=" ++ valType v1 ++ "!")))
			binput -> return (Left (badArgCount optype (L.length binput) 2)))

greaterBuildin :: (String,[RVal] -> IO (Either String RVal))
greaterBuildin = 
	ordBuildin "greater" 

lesserBuildin :: (String,[RVal] -> IO (Either String RVal))
lesserBuildin = 
	ordBuildin "lesser" 
	
eqorgrBuildin :: (String,[RVal] -> IO (Either String RVal))
eqorgrBuildin = 
	ordBuildin "eqorgr" 
	
eqorleBuildin :: (String,[RVal] -> IO (Either String RVal))
eqorleBuildin = 
	ordBuildin "eqorle" 

headBuildin :: (String,[RVal] -> IO (Either String RVal))
headBuildin =
	("head",\input ->
		case input of
			[v] ->
				case v of
					RList ilist ->
						return (Right (L.head ilist))
					_ -> 
						return (Left (badType "head" (deepValType v) "List" "first argument"))
			binput -> return (Left (badArgCount "head" (L.length binput) 1)))
	
tailBuildin :: (String,[RVal] -> IO (Either String RVal))
tailBuildin =
	("tail",\input ->
		case input of
			[v] ->
				case v of
					RList ilist ->
						if ilist == []
						then return (Left (badInput "tail" "the supplied list to be not empty"))
						else return (Right (RList (L.tail ilist)))
					_ -> 
						return (Left (badType "tail" (deepValType v) "List" "first argument"))
			binput -> return (Left (badArgCount "tail" (L.length binput) 1)))
	
initBuildin :: (String,[RVal] -> IO (Either String RVal))
initBuildin =
	("init",\input ->
		case input of
			[v] ->
				case v of
					RList ilist ->
						return (Right (RList (L.init ilist)))
					_ -> 
						return (Left (badType "init" (deepValType v) "List" "first argument"))
			binput -> return (Left (badArgCount "init" (L.length binput) 1)))
	
backBuildin :: (String,[RVal] -> IO (Either String RVal))
backBuildin =
	("back",\input ->
		case input of
			[v] ->
				case v of
					RList ilist ->
						return (Right (L.last ilist))
					_ -> 
						return (Left (badType "back" (deepValType v) "List" "first argument"))
			binput -> return (Left (badArgCount "back" (L.length binput) 1)))
	
growBuildin :: (String,[RVal] -> IO (Either String RVal))
growBuildin =
	("grow",\input ->
		let (_,apb) = appendBuildin in
		apb input)

appendBuildin :: (String,[RVal] -> IO (Either String RVal))
appendBuildin =
	("append",\input ->
		case input of
			(v0:rest) ->
				case valType v0 of
					"RList" ->
						let
							(_,appb) = appendBuildin
							RList list0 = v0
						in do
							appbr <- appb rest
							case appbr of
								Right (RList appbr') -> return (Right (RList (list0 ++ appbr')))
								Left e -> return (Left e)
					_ -> return (Left (badType "append" (fullListTypes input) "[something]" "arguments"))
			[] -> return (Right (RList [])))
	
consBuildin :: (String,[RVal] -> IO (Either String RVal))
consBuildin =
	("cons",\input ->
		case input of
			[v0,v1] ->		
				case valType v1 of
					"RList" ->
						let
							RList ilist = v1 
						in
							if ilist == []
							then return (Right (RList ([v0])))
							else
								case valType v0 == valType (ilist !! 0) of
									True -> return (Right (RList (v0:ilist)))
									False -> 
										return (Left (badInput "cons" 
										("the same type of added instance and list. " ++ "Instance type is " ++ deepValType v0 ++ 
										" and List type is " ++ deepValType v1)))
					err0 -> return (Left (badType "cons" err0 "List" "second argument"))
			binput -> return (Left (badArgCount "cons" (L.length binput) 2)))
			
repeatBuildin :: (String,[RVal] -> IO (Either String RVal))
repeatBuildin =
	("repeat",\input ->
		case input of
			[v,n] ->
				case n of
					RInt rn ->
						return (Right (RList (L.take rn (L.repeat v))))
					_ -> return (Left (badType "repeat" (deepValType n) "Int" "second argument"))
			_ -> return (Left (badArgCount "repeat" (L.length input) 2)))

lenBuildin :: (String,[RVal] -> IO (Either String RVal))
lenBuildin =
	("len",\input -> 
		case input of
			[v] ->
				case v of
					RList l -> return (Right (RInt (L.length l)))
					RTuple l -> return (Right (RInt (L.length l)))
					err -> return (Left (badType "len" (valType err) "List or Tuple" "first argument"))
			binput -> return (Left (badArgCount "len" (L.length binput) 1)))

atBuildin :: (String,[RVal] -> IO (Either String RVal))
atBuildin =
	("at",\input -> 
		case input of
			[v,i'] ->
				case i' of
					RInt i ->	
						case v of
							RList l -> 
								if L.length l <= i
								then
									return (Left (badInput "at" 
									("that the length of List will be greater then entry index! List " ++ show (L.length l) ++ " Index " ++ show i)))
								else
									return (Right (l !! i))
							RTuple l -> 
								if L.length l <= i
								then
									return (Left (badInput "at" 
									("that the length of Tuple will be greater then entry index! Tuple " ++ show (L.length l) ++ " Index " ++ show i)))
								else
									return (Right (l !! i))
							err1 -> return (Left (badType "at" (valType err1) "List or Tuple" "first argument"))
					err -> return (Left (badType "at" (valType err) "Int" "second argument"))
			binput -> return (Left (badArgCount "at" (L.length binput) 2)))

memberBuildin :: (String,[RVal] -> IO (Either String RVal))
memberBuildin =
	("member",\input ->
		case input of
			[k,m] ->		
				case valType m of
					"RMap" -> 
						let 
							RMap imap = m
							isKMember = 
								L.foldl (\a (k',_) -> 
									case valType k == valType k' of
										True ->
											case a of
												Right True -> Right True
												Right False -> Right (k == k') 
												Left err2 -> Left err2
										False -> 
											Left (badInput "member" 
											("keys in map and searched key to be the same type. Searched key type " ++ valType k ++ 
											" and Map key type " ++ valType k' ++ "!")))
								(Right False) imap
						in
							case isKMember of
								Right p -> return (Right (RBool p))
								Left err1 -> return (Left err1)
					err0 -> return (Left (badType "member" err0 "Map" "second argument"))
			binput -> return (Left (badArgCount "member" (L.length binput) 2)))

lookupBuildin :: (String,[RVal] -> IO (Either String RVal))
lookupBuildin =
	("lookup",\input ->
		case input of
			[k,m] ->
				case valType m of
					"RMap" -> 
						let 
							RMap imap = m
							isKMember = 
								L.foldl (\a (k',v') -> 
									case valType k == valType k' of
										True ->
											case a of
												Right RUnit -> if k == k' then Right v' else Right RUnit
												Right foundVal -> Right foundVal 
												Left err2 -> Left err2
										False -> 
											Left (badInput "lookup" 
											("keys in map and searched key to be the same type. Searched key type " ++ valType k ++ 
											" and Map key type " ++ valType k' ++ "!"))) 
								(Right RUnit) imap
						in
							case isKMember of
								Right RUnit -> return (Left (badInput "lookup" ("to find a key " ++ show k ++ " in a map " ++ show m)))
								Right fval -> return (Right fval)
								Left err1 -> return (Left err1)
					err0 -> return (Left (badType "lookup" err0 "Map" "second argument"))
			binput -> return (Left (badArgCount "lookup" (L.length binput) 2)))

insertBuildin :: (String,[RVal] -> IO (Either String RVal))
insertBuildin =
	("insert",\input ->
		case input of
			[k,v,m] ->		
				case valType m of
					"RMap" -> 
						let
							RMap imap = m
							(k',v') = L.head imap 
						in
							if imap == []
							then
								return (Right (RMap [(k,v)]))
							else
								if valType k == valType k'
								then
									if valType v == valType v'
									then
										return (Right (RMap (L.nub $ (k,v) : imap)))
									else 
										return (Left (badInput "insert" 
										("all values to be the same. Inserted key type " ++ valType v ++ " != Map key type " ++ valType v')))
								else 
									return (Left (badInput "insert" 
									("all keys to be the same. Inserted value type " ++ valType k ++ "!= Map val type " ++ valType k')))
					err0 -> return (Left (badType "insert" err0 "Map" "third argument"))
			binput -> return (Left (badArgCount "insert" (L.length binput) 3)))

removeBuildin :: (String,[RVal] -> IO (Either String RVal))
removeBuildin =
	("remove",\input ->
		case input of
			[k,m] ->
				case valType m of
					"RMap" -> 
						let
							RMap imap = m
							(k',_) = L.head imap 
						in
							if imap == []
							then
								return (Right (RMap []))
							else
								if valType k == valType k'
								then
									return (Right (RMap (L.filter (\(k'',_) -> k /= k'') imap)))
								else 
									return (Left (badInput "remove" 
									("all keys to be the same. Removed key type " ++ valType k ++ " != Map key type " ++ valType k')))
					err0 -> return (Left (badType "remove" err0 "Map" "second argument"))
			binput -> return (Left (badArgCount "remove" (L.length binput) 2)))

seqBuildin :: (String,[RVal] -> IO (Either String RVal))
seqBuildin =
	("seq",\_ -> return (Right RUnit))
	
tConstructBuildin :: (String,[RVal] -> IO (Either String RVal))
tConstructBuildin =
	(":t-construct",\input ->
		return (Right (RTuple input)))

lConstructBuildin :: (String,[RVal] -> IO (Either String RVal))
lConstructBuildin =
	(":l-construct",\input ->
		case listType input of
			"error" -> return (Left (badInput "[x0, x1 ... xn]" ("all values for List construction to be the same type")))
			_ -> return (Right (RList input)))

mConstructBuildin :: (String,[RVal] -> IO (Either String RVal))
mConstructBuildin =
	(":m-construct",\input ->
		let
			half [] _ = []
			half (x:xs) p = if p then x:half xs False else half xs True
			keys = half input True 
			vals = half input False
		in
			if L.length keys /= L.length vals
			then
				return (Left (badInput "[k0:v0, k1:v1 ... kn:vn]" 
				("that every key has it's own value (keys " ++ show (L.length keys) ++ " vals " ++ show (L.length vals) ++ ")")))
			else
				if listType keys == "error"
				then
					return (Left (badInput "[k0:v0, k1:v1 ... kn:vn]" "that every key has same type"))	
				else
					if listType vals == "error"
					then
						return (Left (badInput "[k0:v0, k1:v1 ... kn:vn]" "that every value has same type"))	
					else
						return (Right (RMap (L.zip keys vals))))

typeinfoBuildin :: (String,[RVal] -> IO (Either String RVal))
typeinfoBuildin =
	("type-info",\input ->
		case input of
			[v] ->
				return (Right (RList (L.map (\x -> RChar x) (deepValType v))))	
			binput -> return (Left (badArgCount "typeinfo" (L.length binput) 1)))		

typecheckBuildin :: (String,[RVal] -> IO (Either String RVal))
typecheckBuildin =
	("type-check",\input ->
		case input of
			[v] -> 
				if search "error" (deepValType v)
				then 
					return (Left ("(RUNTIME EXCEPTION) Function typecheck detect type error in value " ++ show v ++ 
					". It's type strucutre is " ++ deepValType v ++ ". " ++
					"Error is caused by heterogenous list in this structure!"))
				else 
					return (Right v)	
			binput -> return (Left (badArgCount "typecheck" (L.length binput) 1)))			

errorBuildin :: (String,[RVal] -> IO (Either String RVal))
errorBuildin =
	("error",\input ->
		case input of
			[v] -> 
				case v of
					RList innerlist ->
						case listType innerlist of
							"RChar" ->
								return (Left ("(USER EXCEPTION) " ++ L.map (\(RChar x) -> x) innerlist ++ "!"))
							err1 -> return (Left (badType "error" err1 "Char" "in supplied List"))
					err0 -> return (Left (badType "error" (valType err0) "List" "first argument"))
			binput -> return (Left (badArgCount "error" (L.length binput) 1)))			

traceBuildin :: (String,[RVal] -> IO (Either String RVal))
traceBuildin =
	("trace",\input ->
		case input of
			[m,v] -> 
				case m of
					RList ilist ->
						if listType ilist == "RChar"
						then
							let
								(_,putstrln) = putStrLnBuildin
							in
								putstrln [m] >> return (Right v)
						else return (Left (badType "trace" (listType ilist) "RChar" "list in a place of a first argument"))
					err0 -> return (Left (badType "trace" (valType err0) "List" "first argument"))
			binput -> return (Left (badArgCount "trace" (L.length binput) 2)))	

putStrBuildin :: (String,[RVal] -> IO (Either String RVal))
putStrBuildin =
	("put-str",\input' ->
		IO.hSetBuffering IO.stdout IO.NoBuffering >>
		let (_,putstr) = putStrBuildin in
		case input' of
			[] -> return (Right RUnit)
			(input:rest) ->		
				case unpackString input of
					Just i -> IO.putStr i >> putstr rest
					Nothing -> return (Left (badType "put-str{-ln}" (fullListTypes input') "[[Char],..]" "arguments")))

putStrLnBuildin :: (String,[RVal] -> IO (Either String RVal))
putStrLnBuildin =
	("put-str-ln",\input ->
		let
			(_,putstr) = putStrBuildin
		in
			putstr input >> IO.putStr "\n" >> return (Right RUnit))

readLnBuildin :: (String,[RVal] -> IO (Either String RVal))
readLnBuildin =
	("read-ln",\_ -> 
		IO.getLine >>= (\i -> 
			return (Right (RList (L.map (\c -> RChar c) i)))))

readAllBuildin :: (String,[RVal] -> IO (Either String RVal))
readAllBuildin =
	("read-all",\_ -> 
		IO.getContents >>= (\i -> 
			return (Right (RList (L.map (\c -> RChar c) i)))))

writeFileBuildin :: (String,[RVal] -> IO (Either String RVal))	
writeFileBuildin =
	("write-file",\input ->		
		case input of
			[fname,text] ->
				case unpackString fname of
					Just fraw ->
						case unpackString text of
							Just traw -> 
								IO.writeFile fraw traw >>
								return (Right RUnit)
							Nothing -> return (Left (badType "write-file" (deepValType text) "[Char]" "second argument"))
					Nothing -> return (Left (badType "write-file" (deepValType fname) "[Char]" "first argument"))
			_ -> return (Left (badArgCount "write-file" (L.length input) 2)))

readFileBuildin :: (String,[RVal] -> IO (Either String RVal))	
readFileBuildin =
	("read-file",\input ->		
		case input of
			[fname] ->
				case unpackString fname of
					Just fraw ->
						IO.readFile fraw >>= (\fc ->
							return (Right (RList (L.map (\c -> RChar c) fc))))
					Nothing -> return (Left (badType "read-file" (deepValType fname) "[Char]" "first argument"))
			_ -> return (Left (badArgCount "write-file" (L.length input) 1)))

appendFilesBuildin :: (String,[RVal] -> IO (Either String RVal))			
appendFilesBuildin =
	("append-files",\input ->
		case input of
			(fname:_:frest) ->
				if deepValType fname == "[RChar]"
				then
					if deepListType input
					then
						let
							fpaths = L.map (\(Just x) -> x) (L.map unpackString input) 
						in
							mapM (\fn -> IO.readFile fn) (L.tail fpaths) >>= (\fcs ->
								IO.appendFile (L.head fpaths) (L.concat fcs)) >>
							return (Right RUnit)
					else return (Left (badType "append-file" (fullListTypes input) "[[Char],..]" "arguments"))
				else return (Left (badType "append-file" (fullListTypes input) "[[Char],..]" "arguments"))
			_ -> return (Left (badArgCount "append-files" (L.length input) 2)))

fileExistBuildin :: (String,[RVal] -> IO (Either String RVal))					
fileExistBuildin =
	("file-exist",\input ->
		case input of
			[fname] ->
				case unpackString fname of
					Just fraw ->	
						doesFileExist fraw >>= (\b ->
							return (Right (RBool b)))
					Nothing -> return (Left (badType "file-exist" (deepValType fname) "[Char]" "first argument"))
			_ -> return (Left (badArgCount "file-exist" (L.length input) 1)))		

deleteFilesBuildin :: (String,[RVal] -> IO (Either String RVal))					
deleteFilesBuildin =
	("delete-files",\input ->
		case input of
			(_:_) ->
				if deepValType (input !! 0) == "[RChar]" && deepListType input
				then
					mapM_ (\f -> 
						let 
							Just f' = unpackString f 
						in 
							doesFileExist f' >>= (\p ->
								if p then removeFile f' else return ())) input >> return (Right RUnit)
				else
					return (Left (badType "delete-files" (fullListTypes input) "[[Char],..]" "arguments"))
			_ -> return (Left (badArgCount "delete-files" (L.length input) 1)))		

showBuildin :: (String,[RVal] -> IO (Either String RVal))					
showBuildin =
	("show",\input ->
		case input of
			[val] ->
				return (Right (RList (L.map (\c -> RChar c) (serializeRVal val))))
			_ -> return (Left (badArgCount "show" (L.length input) 1)))	
			
readBuildin :: (String,[RVal] -> IO (Either String RVal))					
readBuildin =
	("read",\input ->
		case input of
			[val] ->
				case unpackString val of
					Just src ->
						return (Right (parseRoseData src))
					Nothing -> return (Left (badType "read" (deepValType val) "[Char]" "first argument"))
			_ -> return (Left (badArgCount "read" (L.length input) 1)))
