module Rose.Eval (
	evalRExpr,
	evalRose,
	evalRoseExpr) where

import System.IO as IO
import Control.Concurrent
import Data.List as L
import Data.Map.Lazy as ML

import Rose.Types
import Rose.Buildins
import Rose.Misc
import Rose.Serialize

evalRExpr :: 
	RExpr 																									{- expression 	-} 
	-> ML.Map String RVal 																	{- outer scope 	-} 
	-> ML.Map String ([RVal] -> IO (Either String RVal)) 		{- buildins 		-} 
	-> IO (Either String RVal)															{- expr result 	-}
evalRExpr (Literal (RLambda lnames lexpr _)) scope _ =
	return (Right (RLambda lnames lexpr scope))
evalRExpr (Literal v) _ _ = 
	return (Right v)
evalRExpr (Var name (REPos fileName lineNum colNum)) scope _ = 
	if ML.member name scope
	then 
		return (Right (scope ! name))
	else
		return (Left ("(RUNTIME EXCEPTION) Variable " ++ name ++ 
		" not found in outer scope! Error location (" ++ fileName ++ 
		":" ++ show lineNum ++ ":" ++ show colNum ++ ")!"))
evalRExpr (Call (Var "par" _) args (REPos fileName lineNum colNum)) scope buildins =
	mapM (\_ -> newEmptyMVar :: IO (MVar (Either String RVal))) args >>= (\mvars ->
		mapM_ (\(v,a) -> forkIO (evalRExpr a scope buildins >>= (\x -> putMVar v x))) (L.zip mvars args) >>
		mapM takeMVar mvars >>= (\evaluatedArgs ->
			if findError evaluatedArgs == ""
			then return (Right (RTuple (collectValues evaluatedArgs)))
			else return (Left (findError evaluatedArgs))))
evalRExpr (Call (Var aname _) args (REPos fileName lineNum colNum)) scope buildins =
	if ML.member aname buildins 
	then
		mapM (\a -> evalRExpr a scope buildins) args >>= (\evaluatedArgs->
			if findError evaluatedArgs == ""
			then 
				(buildins ! aname) (collectValues evaluatedArgs) >>= (\b ->
					case b of
						Right s -> return (Right s)
						Left e -> return (Left (e ++ " At (" ++ fileName ++ ":" ++ show lineNum ++ ":" ++ show colNum ++ ")!")))
			else return (Left (findError evaluatedArgs)))	
	else
		if ML.member aname scope == False
		then
			return (Left ("(RUNTIME EXCEPTION) Function name " ++ 
			aname ++ " not found in outer scope! Error location (" ++ 
			fileName ++ ":" ++ show lineNum ++ ":" ++ show colNum ++ ")!"))
		else
			let
				(RLambda lArgsNames lBody lScope) = scope ! aname
			in
				if L.length args /= L.length lArgsNames
				then
					return (Left ("(RUNTIME EXCEPTION) Invalid number of arguments " ++ 
					"(" ++ show (L.length args) ++ " instead of " ++ 
					show (L.length lArgsNames) ++ ")" ++ " in functuion call in (" ++ 
					fileName ++ ":" ++ show lineNum ++ ":" ++ show colNum ++ ") indirect lambda call from outer scope!"))
				else
					mapM (\a -> evalRExpr a scope buildins) args >>= (\evaluatedArgs->
						if findError evaluatedArgs == ""
						then 
							evalRExpr lBody (ounion (L.foldl (\m (k,v) -> 
								ML.insert k v m) scope (L.zip lArgsNames (collectValues evaluatedArgs))) lScope) buildins
						else 
							return (Left (findError evaluatedArgs)))		
evalRExpr (Call lambdaExpr args (REPos fileName lineNum colNum)) scope buildins =
	evalRExpr lambdaExpr scope buildins >>= (\evaluatedLambda ->
		case evaluatedLambda of
			Right (RLambda lArgsNames lBody lScope) -> 
				if L.length args /= L.length lArgsNames
				then
					return (Left ("(RUNTIME EXCEPTION) Invalid number of arguments " ++ 
					"(" ++ show (L.length args) ++ " instead of " ++ show (L.length lArgsNames) ++ ")" ++ 
					" in functuion call in (" ++ fileName ++ ":" ++ show lineNum ++ ":" ++ show colNum ++ ") indirect lambda call from expression!"))
				else	
					mapM (\a -> evalRExpr a scope buildins) args >>= (\evaluatedArgs->
						if findError evaluatedArgs == ""
						then 
							evalRExpr lBody (ounion (L.foldl (\m (k,v) -> 
								ML.insert k v m) scope (L.zip lArgsNames (collectValues evaluatedArgs))) lScope) buildins
						else return (Left (findError evaluatedArgs)))
			Left err -> return (Left err)
			Right notlambda -> 
				return (Left ("(RUNTIME EXCEPTION) Expression in a place of function call not evaluated to function (" ++ 
				deepValType notlambda ++ " instead of Lambda) " ++ "in (" ++ fileName ++ ":" ++ show lineNum ++ ":" ++ show colNum ++ ")!")))
evalRExpr (LocalBinds binds body) scope buildins =
	let
		valBinds = 
			L.filter (\(name,RLambda anames _ _) -> anames == [":RVALBIND"]) binds
		funBinds = 
			L.filter (\(name,RLambda anames _ _) -> anames /= [":RVALBIND"]) binds
		evalValBinds =
			let 
				exec :: ML.Map String RVal -> [(String,RVal)] -> IO (Either String [(String,RVal)])
				exec pbinds [] =
					return (Right (ML.toList pbinds))
				exec pbinds ((bname,la@(RLambda _ lexpr _)):rest) =
					evalRExpr lexpr (ounion scope pbinds) buildins >>= (\r ->
						case r of
							Right x -> exec (ML.insert bname x pbinds) rest
							Left x -> return (Left x))
			in
				exec ML.empty valBinds 
	in
		evalValBinds >>= (\vb' ->
			case vb' of
				Right vb'' -> 
					let
						vb = vb'' ++ funBinds
						finalizeValBinds = L.map (\(bname,bval) -> 
							case bval of
								RLambda lnames lexpr _ -> (bname,RLambda lnames lexpr (ounion scope (ML.fromList vb))) 
								_ -> (bname,bval)) vb 
					in 
						evalRExpr body (ounion scope (ML.fromList finalizeValBinds)) buildins 
				Left vb'' -> return (Left vb''))	
evalRExpr (Case body branches (REPos fileName lineNum colNum)) scope buildins =
	evalRExpr body scope buildins >>= (\ebody -> 
		case ebody of
			Left err -> return (Left err)
			Right bodyVal -> 
				let
					findBranch [] = 
						Left ("(RUNTIME EXCEPTION) Missing branch in Case expression in (" ++ 
						fileName ++ ":" ++ show lineNum ++ ":" ++ show colNum ++ ")!") 
					findBranch ((RUnit,bexpr):_) =
						Right bexpr
					findBranch ((bval,bexpr):xs) =
						if bval == bodyVal
						then Right bexpr
						else findBranch xs 
				in
					mapM (\a -> evalRExpr a scope buildins) (L.map (\(x,_) -> x) branches) >>= (\evaluatedBranches ->
						if findError evaluatedBranches == ""
						then 
							case findBranch (L.zip (collectValues evaluatedBranches) (L.map (\(_,x) -> x) branches)) of
								Left err -> return (Left err)
								Right branchExpr -> evalRExpr branchExpr scope buildins
						else 
							return (Left (findError evaluatedBranches))))

evalRose :: 
	String																								{- entry function -}
	-> [(String,RVal)]																		{- abstract syntax tree -} 			
	-> ML.Map String ([RVal] -> IO (Either String RVal)) 	{- buildins			 				-}
	-> IO String																					{- result string 				-}
evalRose entry syntacticTree buildins =
		if isMain
		then
			evalRExpr mainExpr scope buildins >>= (\report ->	
				case report of
					Right x -> 
						case x of
							RUnit -> return ""
							_ -> return (serializeRVal x)
					Left err -> return err)
		else return ("(RUNTIME EXCEPTION) " ++ entry ++ " function not found!")
	where
		scope = ML.fromList syntacticTree
		isMain = ML.member entry scope
		RLambda _ mainExpr _ = scope ! entry 

evalRoseExpr ::
	RExpr 																								{- expr for evaluation 									-}
	-> [(String,RVal)]																		{- abstract syntax tree of outer scope 	-}
	-> ML.Map String ([RVal] -> IO (Either String RVal)) 	{- buildins			 												-}
	-> IO String																					{- result string -}
evalRoseExpr expr ast buildins =
	evalRExpr expr (ML.fromList ast) buildins >>= (\report ->
		case report of
			Right x -> 
				case x of
					RUnit -> return ""
					_ -> return (serializeRVal x)
			Left x -> return x)
