module Rose.Graphics.REPL (
	runGraphicsRepl) where

import System.IO as IO
import Data.List as L
import Data.Map.Lazy as ML
import Data.Text as T
import Data.IORef
import Data.Set as S
import Graphics.UI.Gtk
import Control.Concurrent
import Control.Monad
import Control.Monad.Trans
import Control.Exception.Base
import Control.Concurrent.STM
import Control.Monad.STM

import Debug.Trace

import Rose.Eval	
import Rose.Types
import Rose.Parse
import Rose.Serialize
import Rose.Buildins
import Rose.REPL
import Rose.Misc
import Rose.StrTape

runGraphicsRepl :: 
	TextView 																						 -- Text widget for repl
	-> TextBuffer 																			 -- view buffer
	-> TVar [String] 																		 -- files to preinterpret ... aka execution enviroment
	-> String 																					 -- repl start message
	-> [String]																					 -- previosly added buildins names for auto completer
	-> ML.Map String ([RVal] -> IO (Either String RVal)) -- buildins 
	-> IO ()																						 -- Handler which is evaluated when load files are reloaded
	-> IO ()
runGraphicsRepl view buffer loadFiles message pbNames buildins' reloadHandler = do
		ich <- newChan
		pref <- newIORef 4
		history <- newIORef initSTape
		enames <- atomically (newTVar . S.fromList $ (pbNames ++ [":help",":load",":reload",":clear",":exit"]))
		buffer `set` [textBufferText :~ (\x -> x ++ "Welcome in a ROSE REPL v1.0. For help with REPL commands, type :help.\n" ++ 
			if message /= [] then message ++ "\n" else [])]
		on view sizeAllocate (\_ -> do 
			it <- textBufferGetEndIter buffer  
			textViewScrollToIter view it 0 Nothing
			return ())
		on view buttonPressEvent (do
			butt <- eventButton
			if butt == RightButton
			then liftIO (return True)
			else liftIO (return False))
		on view keyPressEvent (do
			code <- eventKeyName
			if T.unpack code == "Return"
			then 
				liftIO (do
					lcount <- textBufferGetLineCount buffer
					ibegin <- textBufferGetIterAtLine buffer (lcount - 1)
					iend <- textBufferGetEndIter buffer
					ltext' <- textBufferGetText buffer ibegin iend True
					let ltext = T.unpack ltext'
					count <- readIORef pref
					writeChan ich (L.drop count ltext)
					submitSTape (writeSTape (L.drop count ltext)) history
					textBufferInsert buffer iend "\n"
					iend <- textBufferGetEndIter buffer
					textBufferPlaceCursor buffer iend
					return True)  
			else if T.unpack code == "BackSpace"
			then do	
				liftIO (do 
					cmark <- textBufferGetInsert buffer
					icursor <- textBufferGetIterAtMark buffer cmark
					cline <- icursor `get` textIterLine
					clineOffset <- icursor `get` textIterLineOffset
					lcount <- textBufferGetLineCount buffer
					ibegin <- textBufferGetIterAtLine buffer (lcount - 1)
					iend <- textBufferGetEndIter buffer
					ltext' <- textBufferGetText buffer ibegin iend True
					count <- readIORef pref
					let ltext = L.reverse (T.unpack ltext')
					if L.take 2 ltext == " >" || L.length ltext <= count || cline < (lcount -1) || clineOffset <= count 
					then return True
					else return False) 
			else if T.unpack code == "Up"
			then do
				liftIO (do
					lcount <- textBufferGetLineCount buffer
					ibegin <- textBufferGetIterAtLine buffer (lcount - 1)
					count <- readIORef pref
					ibegin `set` [textIterOffset :~ (\x -> x + count)]
					iend <- textBufferGetEndIter buffer
					ltext' <- textBufferGetText buffer ibegin iend True
					let ltext = T.unpack ltext'
					submitSTape (shiftBackSTape . writeHeadSTape ltext) history 
					hline <- getSTape headSTape history
					textBufferDeleteInteractive buffer ibegin iend True
					iend <- textBufferGetEndIter buffer
					textBufferInsert buffer iend hline
					return True)
			else if T.unpack code == "Down"
			then do
				liftIO (do
					lcount <- textBufferGetLineCount buffer
					ibegin <- textBufferGetIterAtLine buffer (lcount - 1)
					count <- readIORef pref
					ibegin `set` [textIterOffset :~ (\x -> x + count)]
					iend <- textBufferGetEndIter buffer
					ltext' <- textBufferGetText buffer ibegin iend True
					let ltext = T.unpack ltext'
					submitSTape (shiftForwardSTape . writeHeadSTape ltext) history 
					hline <- getSTape headSTape history
					textBufferDeleteInteractive buffer ibegin iend True
					iend <- textBufferGetEndIter buffer
					textBufferInsert buffer iend hline
					return True)
			else if T.unpack code == "Tab"
			then
				liftIO (do
					cmark <- textBufferGetInsert buffer
					icursor <- textBufferGetIterAtMark buffer cmark
					icursorOffset <- icursor `get` textIterLineOffset
					lcount <- textBufferGetLineCount buffer
					cline <- icursor `get` textIterLine
					if cline < (lcount - 1)
					then return True
					else do
					names <- atomically (readTVar enames)
					let 
						wordsRoot [] = []
						wordsRoot words = 
							if not $ L.foldl (\a w -> if w == [] then False else a) True words
							then []
							else if L.foldl (\a w -> if a == L.head w then a else ' ') (L.head . L.head $ words) (L.tail words) == ' '
							then []
							else (L.head . L.head $ words) : wordsRoot (L.map L.tail words) 
						getword it = 
								getword' it True >>= return . L.reverse
							where
								getword' it isstart = do
									if isstart
									then do
									it `set` [textIterOffset :~ (\x -> x - 1)]
									getword' it False
									else do
									c <- textIterGetChar it
									case c of
										Just ',' ->
											return []
										Just '(' ->
											return []
										Just '=' ->
											return []
										Just ' ' ->
											return [] 
										Just x -> do
											it `set` [textIterOffset :~ (\x -> x - 1)]
											rest <- getword' it False
											return (x : rest) 
										Nothing -> do 
											it `set` [textIterOffset :~ (\x -> x - 1)]
											getword' it False
					tword <- getword icursor
					let pnames = if tword == [] then [] else L.filter (\n -> tword == L.take (L.length tword) n) (S.toList names)
					let strPnames = L.concat . L.intersperse "\n" $ pnames
					case pnames of
						[] -> return ()
						[x] -> do
							if L.head tword == ':'
							then do
								textBufferInsertAtCursor buffer (L.drop (L.length tword) x)
								ilstart <- textBufferGetIterAtLine buffer (lcount - 1)
								ilstart `set` [textIterOffset :~ (\y -> y + icursorOffset + L.length (L.drop (L.length tword) x))]
								textBufferPlaceCursor buffer ilstart
							else do
								textBufferInsertAtCursor buffer (L.drop (L.length tword) x ++ "()")
								lcount <- textBufferGetLineCount buffer
								ilstart <- textBufferGetIterAtLine buffer (lcount - 1)
								ilstart `set` [textIterOffset :~ (\y -> y + icursorOffset + L.length (L.drop (L.length tword) x ++ "("))]
								textBufferPlaceCursor buffer ilstart
						_ ->	
							if L.length (wordsRoot pnames) == (L.length tword) 
							then do
								ibegin <- textBufferGetIterAtLine buffer (lcount - 1)
								iend <- textBufferGetEndIter buffer
								ltext' <- textBufferGetText buffer ibegin iend True
								let ltext = T.unpack ltext'
								textBufferInsert buffer iend ("\n" ++ strPnames ++ "\n" ++ ltext)
								lcount <- textBufferGetLineCount buffer
								ilstart <- textBufferGetIterAtLine buffer (lcount - 1)
								ilstart `set` [textIterOffset :~ (\x -> x + icursorOffset)]
								textBufferPlaceCursor buffer ilstart
							else textBufferInsertAtCursor buffer (L.drop (L.length tword) (wordsRoot pnames)) 				
					return True)
			else if T.unpack code == "Left"
			then return False
			else if T.unpack code == "Right"
			then return False
			else
				liftIO (do 
					lcount <- textBufferGetLineCount buffer
					cmark <- textBufferGetInsert buffer
					icursor <- textBufferGetIterAtMark buffer cmark
					cline <- icursor `get` textIterLine
					clineOffset <- icursor `get` textIterLineOffset
					count <- readIORef pref
					if cline < (lcount - 1) || clineOffset < count
					then return True
					else return False))
		let rthread = (forkFinally (runRepl (readChan ich) (write pref) False loadFiles buildins' reloadHandler 
					(postGUISync $ buffer `set` [textBufferText := ""]) enames) (\e ->
			case e of
				Right _ -> return ()
				Left ex -> 
					case fromException ex of
						Just ThreadKilled -> return ()
						Nothing -> 
							write pref (show ex ++ "\nREPL thread CRASHED! Restarting thread ... OK\n") >> 
							rthread >> 
							return ()))
		rthread >> return ()
	where
		write pref v = do
			modifyIORef pref (\_ -> L.length (L.takeWhile (/='\n') (L.reverse v))) 
			postGUISync (do
				iend <- textBufferGetEndIter buffer
				textBufferInsert buffer iend v)
