module Rose.Misc (
	collectValues,
	findError,
	badType,
	badArgCount,
	badInput,
	search,
	strSlice,
	readMaybe,
	ounion, -- Duplicity se přepisují, nevynechávají jako u union.  
	map2
) where

import Data.List as L
import Data.Map.Lazy as ML

collectValues :: (Show a) => [Either a b] -> [b]
collectValues [] = []
collectValues (Right v:xs) =
	v:collectValues xs
collectValues (Left e:_) =
	error ("collectValues failed thanks to Left -> " ++ show e)

findError :: [Either String b] -> String
findError [] = ""
findError (Right _:xs) = findError xs
findError (Left e:_) = e

badType fun get exp pos =
	"(RUNTIME EXCEPTION) Function " ++ fun ++ " expects " ++ exp ++ " instead of " ++ get ++ (if pos == "" then "" else " in ") ++ pos ++ "!"
	
badArgCount fun get exp =
	"(RUNTIME EXCEPTION) Function " ++ fun ++ " expects " ++ show exp ++ " arguments but it gets " ++ show get ++ "!"
	
badInput fun message =
	"(RUNTIME EXCEPTION) Function " ++ fun ++ " expects " ++ message ++ "!"

search :: String -> String -> Bool
search [] [] = True
search _ [] = False
search [] _ = False
search substr str =
	if L.take (L.length substr) str == substr
	then True
	else search substr (L.tail str)

strSlice :: [String] -> String -> [String]
strSlice _ [] = []
strSlice substr str =
	 	if search "#!@#" str 
	 	then error ("ERROR: strSlice nalezla invalidní sadu charakterů #!@# ve zpracovávaném řetězci: " ++ str)
	 	else L.init (exec str)
	where
		contain [] _ = []
		contain (s:ss) str' =
			if L.take (L.length s) str' == s
			then s
			else contain ss str'
		exec [] = []
		exec str' =
				firstHalf : secondHalf : exec (L.drop (L.length firstHalf + L.length secondHalf) str')
			where
				part [] = []
				part str''@(s:ss) =
					if contain substr str'' == []
					then s : part ss
					else "#!@#" ++ contain substr str''
				(firstHalf, secondHalf) =
					(takeTo "#!@#" (part str'), dropTo "#!@#" (part str'))

dropTo :: String -> String -> String
dropTo _ [] = []
dropTo sub str =
	if L.take (L.length sub) str == sub
	then L.drop (L.length sub) str
	else dropTo sub (L.tail str)
	
takeTo :: String -> String -> String
takeTo _ [] = []
takeTo sub str =
	if L.take (L.length sub) str == sub
	then []
	else L.head str : takeTo sub (L.tail str)

readMaybe :: (Read a) => String -> Maybe a
readMaybe s =
	case [x | (x,t) <- reads s, ("","") <- lex t] of
		[x] -> Just x
		_   -> Nothing

ounion :: (Ord k) => Map k v -> Map k v -> Map k v
ounion f s = L.foldl (\m (k,v) -> ML.insert k v m) f (toList s)

map2 :: (a -> a -> [a]) -> [a] -> [a]
map2 _ [] = []
map2 _ [s] = [s]
map2 f (x:y:xs) = let r = f x y in head r : map2 f (tail r ++ xs)
