module Rose.REPL (
	runRepl) where

import System.IO as IO
import Data.List as L
import Data.Map.Lazy as ML
import Data.IORef
import Data.Set as S
import Text.ParserCombinators.Parsec
import Text.Parsec.Pos
import Text.Parsec.Char
import Control.Concurrent
import Control.Concurrent.STM
import Control.Monad.STM
import Control.Parallel.Strategies

import Rose.Buildins
import Rose.Eval	
import Rose.Types
import Rose.Parse
import Rose.Serialize
import Rose.Misc

parseLoadCommand str =
		case parse parseLoadCommand' "" str of
			Right v -> v
			Left _ -> ""
	where
		parseLoadCommand' :: GenParser Char () String
		parseLoadCommand' =
			skipMany (try space <|> tab) >>
			char ':' >>
			skipMany (try space <|> tab) >>
			(try (string "load") <|> string "l") >>
			skipMany (try space <|> tab) >>
			manyTill anyChar ((newline >> return ()) <|> try (eof >> return ())) >>= (\f -> return f)

myforever io = do
	p <- io 
	if p 
	then return () 
	else myforever io

runRepl ::
	IO String
	-> (String -> IO ())
	-> Bool
	-> TVar [String]
	-> ML.Map String ([RVal] -> IO (Either String RVal))
	-> IO ()
	-> IO ()
	-> TVar (S.Set String)
	-> IO ()
runRepl instream outstream canoff lfiles buildins' rhandler chandler enames = do
	ast <- newIORef [] :: IO (IORef [(String,RVal)])
	atomically (modifyTVar enames (\ns -> L.foldl (flip S.insert) ns ("par" : L.map fst (ML.toList buildins'))))
	rhandler
	lfiles' <- atomically (readTVar lfiles)
	temp <- mapM (\f -> IO.readFile f) lfiles'
	let fasts = parMap rseq (\(fn,fc) -> parseRose fn fc) (L.zip lfiles' temp)
	fasts `seq` mapM_ (\(e,f) -> 
		case e of
			Right ast' -> do
				outstream ("File " ++ f ++ "\t\t\t\t\t\t[OK]\n")
				atomically (modifyTVar enames (\ns -> L.foldl (flip S.insert) ns (L.map fst ast')))
				modifyIORef ast (\x -> L.nub (x ++ ast'))
			Left serrs -> do
				outstream ("File " ++ f ++ "\t\t\t\t\t\t[FAILED]\n") 
				outstream (serrs ++ "\n")) (L.zip fasts lfiles')
	myforever (do
		outstream " <> "
		crline <- instream
		if crline == []
		then return False
		else if parseLoadCommand crline /= "" 
		then do
			input <- IO.readFile (parseLoadCommand crline)
			case parseRose (parseLoadCommand crline) input of
				Right ast' -> do
					outstream ("File " ++ parseLoadCommand crline ++ "\t\t\t\t\t\t[OK]\n")
					atomically (modifyTVar enames (\ns -> L.foldl (flip S.insert) ns (L.map fst ast')))
					modifyIORef ast (\x -> L.nub (x ++ ast')) 
				Left serrs -> do 
					outstream ("File " ++ parseLoadCommand crline ++ "\t\t\t\t\t\t[FAILED]\n") 
					outstream (serrs ++ "\n") 
			return False
		else if crline == ":exit" || crline == ": exit" || crline == ":e"
		then if canoff then return True else outstream "REPL :exit is locked!\n" >> return False
		else if crline == ":help" || crline == ": help" || crline == ":h"
		then do
			outstream (":help -> this help\n" ++
				":load filepath -> loads the specified file into interpreter\n" ++
				":exit -> exit the interpreter, may be locked in some enviroments\n" ++
				":reload -> reloads all loaded files in interpreter\n" ++ 
				":clear -> clear output buffer, it can save many memory when you are working long\n")  
			return False
		else if crline == ":reload" || crline == ":r"
		then do
			rhandler
			lfiles'' <- atomically (readTVar lfiles)
			temp <- mapM (\f -> IO.readFile f) lfiles''
			let fasts = parMap rseq (\(fn,fc) -> parseRose fn fc) (L.zip lfiles'' temp)
			fasts `seq` mapM_ (\(e,f) -> 
				case e of
					Right ast' -> do
						outstream ("File " ++ f ++ "\t\t\t\t\t\t[OK]\n")
						atomically (modifyTVar enames (\ns -> L.foldl (flip S.insert) ns (L.map fst ast')))
						modifyIORef ast (\x -> L.nub (x ++ ast'))
					Left serrs -> do
						outstream ("File " ++ f ++ "\t\t\t\t\t\t[FAILED]\n") 
						outstream (serrs ++ "\n")) (L.zip fasts lfiles'')
			return False
		else if crline == ":clear" || crline == ":c" 
		then chandler >> return False
		else
			case parseRose "console" crline of
				Right ast' -> do 
					atomically (modifyTVar enames (\ns -> L.foldl (flip S.insert) ns (L.map fst ast')))
					modifyIORef ast (\x -> L.nub (x ++ ast')) 
					return False 
				Left serrs -> 
					case parseRoseExpr crline of
						Right e -> do
							cast <- readIORef ast
							evalRoseExpr e cast (buildins' `mergeBuildins` buildins [
								let putstrbuildin = ("put-str",\input' ->
									let (_,putstr) = putstrbuildin in
										case input' of
											[] -> return (Right RUnit)
											(input:rest) ->		
												case unpackString input of
													Just i -> outstream i >> putstr rest
													Nothing -> return (Left (badType "put-str{-ln}" (fullListTypes input') "[[Char],..]" "arguments"))) in putstrbuildin,
								let putstrlnbuildin = ("put-str-ln",\input' ->
									let (_,putstr) = putstrlnbuildin in
										case input' of
											[] -> outstream "\n" >> return (Right RUnit)
											(input:rest) ->		
												case unpackString input of
													Just i -> outstream i >> putstr rest
													Nothing -> return (Left (badType "put-str{-ln}" (fullListTypes input') "[[Char],..]" "arguments"))) in putstrlnbuildin,
								("read-ln",\_ ->
									instream >>= (\i -> 
										return (Right (RList (L.map (\c -> RChar c) i))))),
								("read-all",\_ -> 
									return (Left "(RUNTIME EXCEPTION) Buildin read-all is unsuported in REPL!")),
								("trace",\input -> 	
									case input of
										[m,v] -> 
											case m of
												RList ilist ->
													if listType ilist == "RChar"
													then
														outstream (L.map (\(RChar c) -> c) ilist ++ "\n") >> 
														return (Right v)
													else return (Left (badType "trace" (listType ilist) "RChar" "list in a place of a first argument"))
												err0 -> return (Left (badType "trace" (valType err0) "List" "first argument"))
										binput -> return (Left (badArgCount "trace" (L.length binput) 2)))]) >>= (\o -> do
											if o == ""
											then return ()
											else outstream $ o ++ "\n")
							return False
						Left serrs' -> do
							outstream (serrs ++ "\n\n" ++ serrs' ++ "\n") 
							return False) 
