module Rose.Serialize (
	serializeRVal,
	serializeRExpr,
	serializeFunction,
	serializeRose) where

import Data.List as L
import Data.Map.Lazy as ML

import Rose.Types
import Rose.Misc

serializeRVal :: RVal -> String
serializeRVal (RBool b) = show b
serializeRVal (RInt i) = show i
serializeRVal (RFloat f) = show f
serializeRVal (RChar c) = show c
serializeRVal RUnit = "_"
serializeRVal (RTuple il) = "(" ++ L.concat (L.intersperse "," (L.map serializeRVal il)) ++ ")"
serializeRVal (RList il) = 
	case listType il of
		"RChar" -> "\"" ++ L.map (\(RChar c) -> c) il ++ "\""
		_ -> "[" ++ L.concat (L.intersperse "," (L.map serializeRVal il)) ++ "]"
serializeRVal (RMap il) =
	if il == [] 
	then "[:]"
	else "[" ++ L.concat (L.intersperse "," (L.map (\(f,s) -> serializeRVal f ++ ":" ++ serializeRVal s) il)) ++ "]"
serializeRVal (RLambda anames body _) = 
	"\\(" ++ L.concat (L.intersperse "," anames) ++ ") = " ++ serializeRExpr body ++ ";"

serializeRExpr :: RExpr -> String
serializeRExpr (Literal v) = serializeRVal v
serializeRExpr (Var v _) = v
serializeRExpr (Call (Var ":t-construct" _) args _) = 
	"(" ++ L.concat (L.intersperse "," (L.map serializeRExpr args)) ++ ")"
serializeRExpr (Call (Var ":l-construct" _) args _) = 
	"[" ++ L.concat (L.intersperse "," (L.map serializeRExpr args)) ++ "]"
serializeRExpr (Call (Var ":m-construct" _) args _) = 
	let
		mapArgs [] = []
		mapArgs (x:[]) = []
		mapArgs (x0:x1:rest) = (serializeRExpr x0 ++ ":" ++ serializeRExpr x1) : mapArgs rest 
	in
		"[" ++ L.concat (L.intersperse "," (mapArgs args)) ++ "]"
serializeRExpr (Call fun args _) = 
	serializeRExpr fun ++ "(" ++ L.concat (L.intersperse "," (L.map serializeRExpr args)) ++ ")"
serializeRExpr (LocalBinds binds body) = 
	"{\n"++ unlines (L.map ("\t"++) (lines (L.concat (L.map (\(fname,lambda) -> 
		case lambda of
			RLambda [":RVALBIND"] body _ -> fname ++ " = " ++ serializeRExpr body ++ ";\n"
			_ -> serializeFunction fname lambda ++ "\n") binds)))) ++ "} " ++ serializeRExpr body
serializeRExpr (Case body branches _) =
	serializeRExpr body ++ " {\n"++ unlines (L.map ("\t"++) (lines (L.concat (L.map (\(e0,e1) -> 
		serializeRExpr e0 ++ " : " ++ serializeRExpr e1 ++ ";\n") branches)))) ++ "}"

serializeFunction :: String -> RVal -> String
serializeFunction fname (RLambda anames body _) = 
	fname ++ "(" ++ L.concat (L.intersperse "," anames) ++ ")" ++ " = " ++ serializeRExpr body ++ ";"

serializeRose :: [(String,RVal)] -> String
serializeRose [] = ""
serializeRose ((fname,fbody):rest) = serializeFunction fname fbody ++ "\n" ++ serializeRose rest
